# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .arithmetic import (
    add, sub, mod, mul, mul_prepared, prepare_for_mul, prepare_for_mul_cpu, pow, inv_pow2, lsh)
from .ntt import ntt1024, ntt1024_requirements
from .fft import fft512, fft512_requirements
from .computation import Transform



import random
import numpy
import nufhe
from reikna.cluda import any_api

size = 32
bits1 = [random.choice([False, True]) for i in range(size)]
bits2 = [random.choice([False, True]) for i in range(size)]
reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

thr = any_api().Thread.create(interactive=True)

rng = nufhe.DeterministicRNG()
secret_key, cloud_key = nufhe.make_key_pair(thr, rng)

ciphertext1 = nufhe.encrypt(thr, rng, secret_key, bits1)
ciphertext2 = nufhe.encrypt(thr, rng, secret_key, bits2)

result = nufhe.empty_ciphertext(thr, cloud_key.params, ciphertext1.shape)
nufhe.gate_nand(thr, cloud_key, result, ciphertext1, ciphertext2)

result_bits = nufhe.decrypt(thr, secret_key, result)

assert all(result_bits == reference)




# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
LWE (Learning With Errors) functions.
"""

import io
import pickle

import numpy

from reikna.cluda.api import Thread
from reikna.core import Type
import reikna

from .utils import arrays_equal
from .numeric_functions import (
    Torus32,
    ErrorFloat,
    )
from .lwe_gpu import (
    LweKeyswitch,
    MakeLweKeyswitchKey,
    LweEncrypt,
    LweDecrypt,
    LweLinear,
    LweNoiselessTrivial,
    LweNoiselessTrivialConstant,
    )
from .random_numbers import (
    rand_uniform_bool,
    rand_uniform_torus32,
    rand_gaussian_torus32,
    )
from .computation_cache import get_computation


class LweParams:

    def __init__(self, size: int, min_noise: float, max_noise: float):
        self.size = size
        self.min_noise = min_noise # the smallest noise that makes it secure
        self.max_noise = max_noise # the biggest noise that allows decryption

    def __eq__(self, other: 'LweParams'):
        return (
            self.__class__ == other.__class__
            and self.size == other.size
            and self.min_noise == other.min_noise
            and self.max_noise == other.max_noise)

    def __hash__(self):
        return hash((self.__class__, self.size, self.min_noise, self.max_noise))


class LweKey:

    def __init__(self, params: LweParams, key):
        self.params = params
        self.key = key

    @classmethod
    def from_rng(cls, thr: Thread, params: LweParams, rng):
        return cls(params, rand_uniform_bool(thr, rng, (params.size,)))

    # extractions ring Lwe * Lwe
    @classmethod
    def from_tlwe_key(cls, params: LweParams, tlwe_key: 'TLweKey'):
        poly_degree = tlwe_key.params.polynomial_degree
        mask_size = tlwe_key.params.mask_size
        assert params.size == poly_degree * mask_size

        key = tlwe_key.key.coeffs.ravel()

        return cls(params, key)

    def dump(self, file_obj):
        pickle.dump(self.params, file_obj)
        pickle.dump(self.key.get(), file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        params = pickle.load(file_obj)
        key = pickle.load(file_obj)
        return cls(params, thr.to_device(key))

    def __eq__(self, other: 'LweKey'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and arrays_equal(self.key, other.key))


class LweSampleArrayShapeInfo:

    def __init__(self, a, b, current_variances):

        if (not (len(a.shape) - 1 == len(b.shape) == len(current_variances.shape))
                or not (a.shape[:-1] == b.shape == current_variances.shape)):

            raise ValueError("Inconsistent shapes: {a}, {b}, {cv}".format(
                a=a.shape, b=b.shape, cv=current_variances.shape))

        self.a = Type.from_value(a)
        self.b = Type.from_value(b)
        self.current_variances = Type.from_value(current_variances)
        self.shape = b.shape

    def __eq__(self, other: 'LweSampleArrayShapeInfo'):
        return (
            self.__class__ == other.__class__
            and self.a == other.a
            and self.b == other.b
            and self.current_variances == other.current_variances)

    def __hash__(self):
        return hash((self.__class__, self.a, self.b, self.current_variances))


class LweSampleArray:
    """
    A ciphertext object.

    .. py:attribute:: shape

        The shape of the encrypted plaintext message.
    """

    def __init__(self, params: LweParams, a, b, current_variances):
        """__init__()""" # hide the signature from Sphinx
        self.params = params
        self.a = a
        self.b = b
        self.current_variances = current_variances
        self.shape_info = LweSampleArrayShapeInfo(a, b, current_variances)

    @classmethod
    def empty(cls, thr: Thread, params: LweParams, shape):
        a = thr.array(shape + (params.size,), Torus32)
        b = thr.array(shape, Torus32)
        current_variances = thr.array(shape, ErrorFloat)
        return cls(params, a, b, current_variances)

    @property
    def shape(self):
        return self.shape_info.shape

    def __getitem__(self, index):
        """
        Returns a view over the ciphertext (still a :py:class:`LweSampleArray` object).
        The indexing works in the same way as if it was a regular ``numpy`` array
        with the shape ``shape``.
        """
        a_view = self.a[index]
        b_view = self.b[index]
        cv_view = self.current_variances[index]
        return LweSampleArray(self.params, a_view, b_view, cv_view)

    def __setitem__(self, index, value):
        if not isinstance(value, LweSampleArray):
            raise ValueError("Only assignment of ciphertexts is supported")
        self.a[index] = value.a
        self.b[index] = value.b
        self.current_variances[index] = value.current_variances

    def copy(self):
        """
        Returns a copy of the ciphertext.
        """
        return LweSampleArray(
            self.params, self.a.copy(), self.b.copy(), self.current_variances.copy())

    def roll(self, shift, axis=-1):
        """
        Cyclically shifts encrypted bits of the cyphertext **inplace**
        by ``shift`` positions to the right along ``axis``.
        ``shift`` can be negative (in which case the elements are shifted to the left).
        Elements that are shifted beyond the last position are re-introduced at the first
        (and vice versa).

        Works equivalently to ``numpy.roll`` (except ``axis=None`` is not supported).
        """
        if shift == 0:
            return

        axis = axis % len(self.shape)

        self.a.roll(shift, axis=axis)
        self.b.roll(shift, axis=axis)
        self.current_variances.roll(shift, axis=axis)

    def dump(self, file_obj):
        """
        Serialize into the given ``file_obj``, a writeable file-like object.
        """
        pickle.dump(self.params, file_obj)
        pickle.dump(self.a.get(), file_obj)
        pickle.dump(self.b.get(), file_obj)
        pickle.dump(self.current_variances.get(), file_obj)

    def dumps(self):
        """
        Serialize into a bytestring.
        """
        file_obj = io.BytesIO()
        self.dump(file_obj)
        return file_obj.getvalue()

    @classmethod
    def load(cls, file_obj, thr):
        """
        Deserialize from the given ``file_obj``, a readable file-like object,
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        params = pickle.load(file_obj)
        a = thr.to_device(pickle.load(file_obj))
        b = thr.to_device(pickle.load(file_obj))
        current_variances = thr.to_device(pickle.load(file_obj))
        return cls(params, a, b, current_variances)

    @classmethod
    def loads(cls, s, thr):
        """
        Deserialize from the given bytestring
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        file_obj = io.BytesIO(s)
        return cls.load(file_obj, thr)

    def __eq__(self, other: 'LweSampleArray'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and arrays_equal(self.a, other.a)
            and arrays_equal(self.b, other.b)
            and arrays_equal(self.current_variances, other.current_variances))


class LweKeyswitchKey:

    def __init__(self, lwe: LweSampleArray):
        input_size, decomp_length, base = lwe.shape

        self.lwe = lwe
        self.input_size = input_size # length of the input key: s'
        self.output_size = lwe.params.size # params of the output key s
        self.decomp_length = decomp_length # decomposition length
        self.log2_base = int(numpy.log2(base)) # log_2(decomposition base)

    @classmethod
    def from_tgsw_key(
            cls, thr, rng, ks_decomp_length: int, ks_log2_base: int,
            lwe_key: LweKey, tgsw_key: 'TGswKey'):

        bk_params = tgsw_key.params
        accum_params = bk_params.tlwe_params
        extract_params = accum_params.extracted_lweparams
        extracted_key = LweKey.from_tlwe_key(extract_params, tgsw_key.tlwe_key)

        in_key = extracted_key
        out_key = lwe_key

        input_size = in_key.params.size
        output_size = out_key.params.size
        noise = out_key.params.min_noise
        base = 2**ks_log2_base

        lwe = LweSampleArray.empty(thr, out_key.params, (input_size, ks_decomp_length, base))

        noises_b = rand_gaussian_torus32(
            thr, rng, 0, noise, (input_size, ks_decomp_length, base - 1), centered=True)
        noises_a = rand_uniform_torus32(
            thr, rng, (input_size, ks_decomp_length, base - 1, output_size))

        comp = get_computation(
            thr, MakeLweKeyswitchKey,
            input_size, output_size, ks_decomp_length, ks_log2_base, noise)
        comp(lwe.a, lwe.b, lwe.current_variances, in_key.key, out_key.key, noises_a, noises_b)

        return cls(lwe)

    def dump(self, file_obj):
        self.lwe.dump(file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        lwe = LweSampleArray.load(file_obj, thr)
        return cls(lwe)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.lwe == other.lwe)


def lwe_keyswitch(thr: Thread, result: LweSampleArray, ks: LweKeyswitchKey, sample: LweSampleArray):
    """
    Translate the message of the result sample by -sum(a[i].s[i]) where s is the secret.
    """
    lwe = ks.lwe
    comp = get_computation(
        thr, LweKeyswitch,
        result.shape_info, ks.input_size, ks.output_size, ks.decomp_length, ks.log2_base)
    # Note: sample.current_variances are ignored.
    comp(
        result.a, result.b, result.current_variances,
        lwe.a, lwe.b, lwe.current_variances, sample.a, sample.b)


def lwe_encrypt(thr: Thread, rng, result: LweSampleArray, messages, noise: float, key: LweKey):
    """
    Encrypt a message with the secret key, with stdev `noise`.
    """
    lwe_size = key.params.size
    noises_b = rand_gaussian_torus32(thr, rng, 0, noise, messages.shape)
    noises_a = rand_uniform_torus32(thr, rng, messages.shape + (lwe_size,))
    comp = get_computation(thr, LweEncrypt, messages.shape, lwe_size, noise)
    comp(result.a, result.b, result.current_variances, messages, key.key, noises_a, noises_b)


def lwe_decrypt(thr: Thread, sample: LweSampleArray, key: LweKey):
    """
    Decrypt an LWE array with the secret key.
    """
    result = thr.empty_like(sample.b)
    comp = get_computation(thr, LweDecrypt, sample.shape_info.shape, key.params.size)
    comp(result, sample.a, sample.b, key.key)
    return result.get()


def lwe_noiseless_trivial(thr: Thread, result: LweSampleArray, mus):
    """
    Initialize LWE samples with `(0, mu)` for each `mu` in `mus`.
    """
    comp = get_computation(thr, LweNoiselessTrivial, result.shape_info, mus.shape)
    comp(result.a, result.b, result.current_variances, mus)


def lwe_noiseless_trivial_constant(thr: Thread, result: LweSampleArray, mu):
    """
    Initialize LWE samples with `(0, mu)`.
    """
    comp = get_computation(thr, LweNoiselessTrivialConstant, result.shape_info)
    comp(result.a, result.b, result.current_variances, mu)


# Arithmetic operations on LWE samples


def lwe_negate(thr: Thread, result: LweSampleArray, source: LweSampleArray):
    """
    result = -sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, -1)


def lwe_copy(thr: Thread, result: LweSampleArray, source: LweSampleArray):
    """
    result = sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, 1)


def lwe_add_to(thr: Thread, result: LweSampleArray, source: LweSampleArray):
    """
    result += sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info, add_result=True)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, 1)


def lwe_add_mul_to(thr: Thread, result: LweSampleArray, p: int, source: LweSampleArray):
    """
    result += p * sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info, add_result=True)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, p)


def lwe_sub_to(thr: Thread, result: LweSampleArray, source: LweSampleArray):
    """
    result -= sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info, add_result=True)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, -1)


def lwe_sub_mul_to(thr: Thread, result: LweSampleArray, p: int, source: LweSampleArray):
    """
    result -= p * sample
    """
    comp = get_computation(thr, LweLinear, result.shape_info, source.shape_info, add_result=True)
    comp(
        result.a, result.b, result.current_variances,
        source.a, source.b, source.current_variances, -p)


def concatenate(lwe_sample_arrays, axis=0, out=None):
    """
    Concatenates several ciphertext arrays along ``axis``.
    """
    if len(lwe_sample_arrays) == 0:
        raise ValueError("Need at least one ciphertext to concatenate")

    params = lwe_sample_arrays[0].params
    lwes_a = [lwe.a for lwe in lwe_sample_arrays]
    lwes_b = [lwe.b for lwe in lwe_sample_arrays]
    lwes_cv = [lwe.current_variances for lwe in lwe_sample_arrays]
    if out is None:
        out = LweSampleArray(
            params,
            reikna.concatenate(lwes_a, axis=axis),
            reikna.concatenate(lwes_b, axis=axis),
            reikna.concatenate(lwes_cv, axis=axis))
    else:
        reikna.concatenate(lwes_a, axis=axis, out=out.a)
        reikna.concatenate(lwes_b, axis=axis, out=out.b)
        reikna.concatenate(lwes_cv, axis=axis, out=out.current_variances)

    return out



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

import reikna.helpers as helpers
from reikna.cluda import OutOfResourcesError
from reikna.core import Computation, Parameter, Annotation, Type


TEMPLATE = helpers.template_for(__file__)


class Transform(Computation):

    def __init__(
            self, transform, batch_shape, inverse=False,
            i32_conversion=False, transforms_per_block=4, kernel_repetitions=1):

        self._inverse = inverse
        self._transform = transform
        self._transforms_per_block = transforms_per_block
        self._kernel_repetitions = kernel_repetitions
        self._i32_conversion = i32_conversion

        tr_arr = Type(self._transform.elem_dtype, batch_shape + (transform.transform_length,))
        if i32_conversion:
            arr = Type(numpy.int32, batch_shape + (transform.polynomial_length,))
            if inverse:
                oarr = arr
                iarr = tr_arr
            else:
                oarr = tr_arr
                iarr = arr
        else:
            oarr = tr_arr
            iarr = tr_arr

        Computation.__init__(self, [
            Parameter('output', Annotation(oarr, 'o')),
            Parameter('input', Annotation(iarr, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):

        plan = plan_factory()

        batch_size = helpers.product(output.shape[:-1])

        cdata_arr = self._transform.cdata_inv if self._inverse else self._transform.cdata_fw
        if self._transform.use_constant_memory:
            cdata = plan.constant_array(cdata_arr)
        else:
            cdata = plan.persistent_array(cdata_arr)

        tpb = self._transforms_per_block
        while tpb >= 1:
            blocks_num = helpers.min_blocks(batch_size, tpb)
            try:
                plan.kernel_call(
                    TEMPLATE.get_def('standalone_transform'),
                        [output, input_, cdata],
                        kernel_name="standalone_transform",
                        global_size=(
                            blocks_num,
                            self._transform.threads_per_transform * tpb),
                        local_size=(
                            1,
                            self._transform.threads_per_transform * tpb),
                        render_kwds=dict(
                            inverse=self._inverse,
                            i32_conversion=self._i32_conversion,
                            kernel_repetitions=self._kernel_repetitions,
                            transform=self._transform,
                            transforms_per_block=tpb,
                            batch_size=batch_size,
                            blocks_num=blocks_num,
                            slices=(len(output.shape) - 1, 1)))
                break
            except OutOfResourcesError:
                tpb -= 1
        else:
            raise Exception(
                "The selected device does not have enough resources for the selected transform")

        return plan



import os
import random
import nufhe


def client_prepare():

    size = 32
    bits1 = [random.choice([False, True]) for i in range(size)]
    bits2 = [random.choice([False, True]) for i in range(size)]
    reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

    ctx = nufhe.Context()
    secret_key, cloud_key = ctx.make_key_pair()
    ciphertext1 = ctx.encrypt(secret_key, bits1)
    ciphertext2 = ctx.encrypt(secret_key, bits2)

    with open('secret_key', 'wb') as f:
        secret_key.dump(f)

    with open('cloud_key', 'wb') as f:
        cloud_key.dump(f)

    with open('ciphertext1', 'wb') as f:
        ciphertext1.dump(f)

    with open('ciphertext2', 'wb') as f:
        ciphertext2.dump(f)

    return reference


def cloud_process():

    ctx = nufhe.Context()

    with open('cloud_key', 'rb') as f:
        cloud_key = ctx.load_cloud_key(f)

    vm = ctx.make_virtual_machine(cloud_key)

    with open('ciphertext1', 'rb') as f:
        ciphertext1 = vm.load_ciphertext(f)

    with open('ciphertext2', 'rb') as f:
        ciphertext2 = vm.load_ciphertext(f)

    result = vm.gate_nand(ciphertext1, ciphertext2)

    with open('result', 'wb') as f:
        result.dump(f)


def client_verify(reference):

    ctx = nufhe.Context()

    with open('secret_key', 'rb') as f:
        secret_key = ctx.load_secret_key(f)

    with open('result', 'rb') as f:
        result = ctx.load_ciphertext(f)

    result_bits = ctx.decrypt(secret_key, result)
    assert all(result_bits == reference)


def cleanup():
    os.remove('secret_key')
    os.remove('cloud_key')
    os.remove('ciphertext1')
    os.remove('ciphertext2')
    os.remove('result')


reference = client_prepare()
cloud_process()
client_verify(reference)
cleanup()



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.helpers import product

from .numeric_functions import Torus32
from .polynomial_transform import get_transform


def TLweNoiselessTrivialReference(params: 'TLweParams', shape):

    mask_size = params.mask_size
    polynomial_degree = params.polynomial_degree
    batch_len = product(shape)

    def _kernel(a, current_variances, mu):
        a_view = a.reshape(batch_len, mask_size + 1, polynomial_degree)
        a_view[:,:mask_size,:] = 0
        a_view[:,mask_size,:] = mu.reshape(batch_len, polynomial_degree)
        current_variances.fill(0)

    return _kernel


def TLweExtractLweSamplesReference(params: 'TLweParams', shape):

    mask_size = params.mask_size
    polynomial_degree = params.polynomial_degree
    batch_len = product(shape)

    def _kernel(result_a, result_b, tlwe_a):

        batch = product(tlwe_a.shape[:-2])

        a_view = result_a.reshape(batch_len, mask_size, polynomial_degree)
        b_view = result_b.reshape(batch_len)
        tlwe_a_view = tlwe_a.reshape(batch_len, mask_size + 1, polynomial_degree)

        a_view[:,:,0] = tlwe_a_view[:, :mask_size, 0]
        a_view[:,:,1:] = -tlwe_a_view[:, :mask_size, :0:-1]

        numpy.copyto(b_view, tlwe_a_view[:, mask_size, 0])

    return _kernel


# create an homogeneous tlwe sample
def TLweEncryptZeroReference(params: 'TLweParams', shape, noise: float, perf_params):
    polynomial_degree = params.polynomial_degree
    mask_size = params.mask_size

    transform = get_transform(params.transform_type)
    tr_dtype = transform.transformed_dtype()

    def _kernel(result_a, result_cv, key, noises1, noises2):
        batch_len = product(shape)
        result_a_view = result_a.reshape(batch_len, mask_size + 1, polynomial_degree)
        noises1_view = noises1.reshape(batch_len, mask_size, polynomial_degree)
        noises2_view = noises2.reshape(batch_len, polynomial_degree)

        tmp1 = transform.forward_transform_ref(key)
        tmp2 = transform.forward_transform_ref(noises1_view)
        tmp3 = transform.transformed_space_mul_ref(tmp1, tmp2)
        tmpr = transform.inverse_transform_ref(tmp3)

        result_a_view[:,:mask_size,:] = noises1_view
        result_a_view[:,mask_size,:] = noises2_view
        for i in range(mask_size):
            result_a_view[:,mask_size,:] += tmpr[:,i,:]

        result_cv.fill(noise**2)

    return _kernel



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy
import pytest

import nufhe.transform.ntt_cpu as ntt_cpu


def tp_dtype(tp):
    if isinstance(tp, str) and tp == 'ff_number':
        return numpy.uint64
    else:
        return tp


def tp_limits(tp):
    if isinstance(tp, str) and tp == 'ff_number':
        return 0, ntt_cpu.GaloisNumber.modulus
    elif numpy.issubdtype(tp, numpy.integer):
        ii = numpy.iinfo(tp)
        return ii.min, ii.max + 1
    else:
        return -10, 10


def get_test_array(shape, tp, val_range=None):
    dtype = tp_dtype(tp)
    if val_range is None:
        nmin, nmax = tp_limits(tp)
    else:
        nmin, nmax = val_range

    if numpy.issubdtype(dtype, numpy.integer):
        return numpy.random.randint(nmin, nmax, dtype=dtype, size=shape)
    elif numpy.issubdtype(dtype, numpy.floating):
        return numpy.random.uniform(nmin, nmax, size=shape).astype(dtype)
    elif numpy.issubdtype(dtype, numpy.complexfloating):
        return (
            numpy.random.uniform(nmin, nmax, size=shape)
            + 1j * numpy.random.uniform(nmin, nmax, size=shape)).astype(dtype)
    else:
        raise NotImplementedError(dtype)


def errors_allclose(arr1, arr2):
    # Errors are single-precision floats, and implementations of single-float arithmetic
    # tend to give slightly different answer depending on the device.
    # So the default values of `numpy.allclose` are too strict.
    return numpy.allclose(arr1, arr2, rtol=1e-3)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time

import numpy

from .numeric_functions import phase_to_t32
from .lwe import (
    LweSampleArray,
    lwe_keyswitch,
    )
from .lwe import (
    lwe_add_to,
    lwe_sub_to,
    lwe_add_mul_to,
    lwe_sub_mul_to,
    lwe_noiseless_trivial,
    lwe_noiseless_trivial_constant,
    lwe_negate,
    lwe_copy,
    )
from .api_low_level import NuFHECloudKey, bool_to_t32
from .bootstrap import bootstrap
from .performance import PerformanceParameters, PerformanceParametersForDevice


def get_shape(obj):
    if hasattr(obj, 'shape'):
        return obj.shape
    elif isinstance(obj, list):
        return numpy.asarray(obj).shape
    else:
        raise ValueError("An object of type " + str(type(obj)) + " is not array-like")


def _result_shape_pair(shape1, shape2):
    if len(shape1) > len(shape2):
        shape2 = (1,) * (len(shape1) - len(shape2)) + shape2
    else:
        shape1 = (1,) * (len(shape2) - len(shape1)) + shape1

    if any((l1 != l2 and l1 > 1 and l2 > 1) for l1, l2 in zip(shape1, shape2)):
        raise ValueError("Incompatible shapes: {s1}, {s2}".format(s1=shape1, s2=shape2))

    return tuple((l1 if l1 > 1 else l2) for l1, l2 in zip(shape1, shape2))


def result_shape(*shapes):
    if len(shapes) == 1:
        return shapes[0]
    elif len(shapes) == 2:
        return _result_shape_pair(*shapes)
    else:
        return _result_shape_pair(shapes[0], result_shape(*shapes[1:]))


def check_shape(result, *args):
    rshape = result_shape(*[arg.shape for arg in args])
    if len(rshape) > len(result.shape) or rshape != result.shape[len(result.shape)-len(rshape):]:
        raise ValueError(
            ("The shape of the result derived from the arguments {derived_shape} "
            "cannot be broadcasted to the shape of the destination {dest_shape}").format(
            derived_shape=rshape, dest_shape=result.shape))


def gate_nand(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped NAND gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params

    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,1/8) - a - b
    NandConst = phase_to_t32(1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, NandConst)
    lwe_sub_to(thr, temp_result, a)
    lwe_sub_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_or(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped OR gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,1/8) + a + b
    OrConst = phase_to_t32(1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, OrConst)
    lwe_add_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_and(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped AND gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,-1/8) + a + b
    AndConst = phase_to_t32(-1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, AndConst)
    lwe_add_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_xor(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped XOR gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,1/4) + 2*(a + b)
    XorConst = phase_to_t32(1, 4)
    lwe_noiseless_trivial_constant(thr, temp_result, XorConst)
    lwe_add_mul_to(thr, temp_result, 2, a)
    lwe_add_mul_to(thr, temp_result, 2, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_xnor(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped XNOR gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,-1/4) + 2*(-a-b)
    XnorConst = phase_to_t32(-1, 4)
    lwe_noiseless_trivial_constant(thr, temp_result, XnorConst)
    lwe_sub_mul_to(thr, temp_result, 2, a)
    lwe_sub_mul_to(thr, temp_result, 2, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_not(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray, a: LweSampleArray,
        perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic NOT gate.
    Applied elementwise on an encrypted array of bits.

    Not bootstrapped; ``perf_params`` does not have any effect and is only present
    for the sake of API uniformity.

    The shape of ``a`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the source ciphertext.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 1 LWE sample (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE sample (with message space [-1/8,1/8], noise<1/16)

    check_shape(result, a)

    in_out_params = cloud_key.params.in_out_params
    lwe_negate(thr, result, a)


def gate_copy(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray, a: LweSampleArray,
        perf_params: PerformanceParametersForDevice=None):
    """
    Copy the contents of the ciphertext ``a`` to ``result``.

    Not bootstrapped; ``perf_params`` does not have any effect and is only present
    for the sake of API uniformity.

    The shape of ``a`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the source ciphertext.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 1 LWE sample (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE sample (with message space [-1/8,1/8], noise<1/16)

    check_shape(result, a)

    in_out_params = cloud_key.params.in_out_params
    lwe_copy(thr, result, a)


"""
 * Homomorphic Trivial Constant gate (doesn't need to be bootstrapped)
 * Takes a boolean value)
 * Outputs a LWE sample (with message space [-1/8,1/8], noise<1/16)
"""
def gate_constant(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray, vals,
        perf_params: PerformanceParametersForDevice=None):
    """
    Fill each bit of the ciphertext ``result`` with the trivial encryption
    of the plaintext values from ``vals`` (which will be converted to ``bool``).

    ``vals`` should be an array or a list with a shape broadcastable to the shape of ``result``,
    or a scalar value.

    .. note::

        "Trivial encryption" means that the result of this gate does not require
        a secret key for decryption, and cannot be used to implement public key encryption.
        Its intended purpose is to initialize constants in bootstrapped circuits.

    Not bootstrapped; ``perf_params`` does not have any effect and is only present
    for the sake of API uniformity.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
        Must be the same shape as the ``vals`` array.
    :param vals: a ``numpy.bool`` array (or anything castable to it) used to fill the ciphertext.
    :param perf_params: an override for performance parameters.
    """

    vals = numpy.asarray(vals)
    vals = bool_to_t32(vals)

    check_shape(result, vals)

    if vals.ndim == 0:
        lwe_noiseless_trivial_constant(thr, result, vals)
    else:
        lwe_noiseless_trivial(thr, result, thr.to_device(vals))


def gate_nor(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped NOR gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,-1/8) - a - b
    NorConst = phase_to_t32(-1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, NorConst)
    lwe_sub_to(thr, temp_result, a)
    lwe_sub_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_andny(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped ANDNY (`(not a) and b`) gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,-1/8) - a + b
    AndNYConst = phase_to_t32(-1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, AndNYConst)
    lwe_sub_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_andyn(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped ANDYN (`a and (not b)`) gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,-1/8) + a - b
    AndYNConst = phase_to_t32(-1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, AndYNConst)
    lwe_add_to(thr, temp_result, a)
    lwe_sub_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_orny(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped ORNY (`(not a) or b`) gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,1/8) - a + b
    OrNYConst = phase_to_t32(1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, OrNYConst)
    lwe_sub_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_oryn(
        thr, cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped ORYN (`a or (not b)`) gate.
    Applied elementwise on two encrypted arrays of bits.

    The shapes of ``a`` and ``b`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 2 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b)

    in_out_params = cloud_key.params.in_out_params
    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)

    #compute: (0,1/8) + a - b
    OrYNConst = phase_to_t32(1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, OrYNConst)
    lwe_add_to(thr, temp_result, a)
    lwe_sub_to(thr, temp_result, b)

    #if the phase is positive, the result is 1/8
    #if the phase is positive, else the result is -1/8
    MU = phase_to_t32(1, 8)
    bootstrap(
        thr, result, cloud_key.bootstrap_key, cloud_key.keyswitch_key,
        MU, temp_result, perf_params)


def gate_mux(
        thr,
        cloud_key: NuFHECloudKey, result: LweSampleArray,
        a: LweSampleArray, b: LweSampleArray, c: LweSampleArray,
        perf_params: PerformanceParametersForDevice=None):
    """
    Homomorphic bootstrapped MUX (`b if a else c`, or, equivalently,
    `(a and b) or ((not a) and c)`) gate.
    Applied elementwise on three encrypted arrays of bits.

    The shapes of ``a``, ``b`` and ``c`` should be broadcastable to the shape of ``result``.

    :param thr: a ``reikna`` ``Thread`` object.
    :param cloud_key: the cloud key.
    :param result: an empty ciphertext where the result will be stored.
    :param a: the ciphertext with the first argument.
    :param b: the ciphertext with the second argument.
    :param c: the ciphertext with the third argument.
    :param perf_params: an override for performance parameters.
    """

    # * Takes in input 3 LWE samples (with message space [-1/8,1/8], noise<1/16)
    # * Outputs a LWE bootstrapped sample (with message space [-1/8,1/8], noise<1/16)

    if perf_params is None:
        perf_params = PerformanceParameters(cloud_key.params).for_device(thr.device_params)

    check_shape(result, a, b, c)

    MU = phase_to_t32(1, 8)
    in_out_params = cloud_key.params.in_out_params
    extracted_params = cloud_key.params.tgsw_params.tlwe_params.extracted_lweparams

    temp_result = LweSampleArray.empty(thr, in_out_params, result.shape)
    temp_result1 = LweSampleArray.empty(thr, extracted_params, result.shape)
    u1 = LweSampleArray.empty(thr, extracted_params, result.shape)
    u2 = LweSampleArray.empty(thr, extracted_params, result.shape)

    #compute "AND(a,b)": (0,-1/8) + a + b
    AndConst = phase_to_t32(-1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result, AndConst)
    lwe_add_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, b)
    # Bootstrap without KeySwitch
    bootstrap(
        thr, u1, cloud_key.bootstrap_key, cloud_key.keyswitch_key, MU, temp_result,
        perf_params, no_keyswitch=True)

    #compute "AND(not(a),c)": (0,-1/8) - a + c
    lwe_noiseless_trivial_constant(thr, temp_result, AndConst)
    lwe_sub_to(thr, temp_result, a)
    lwe_add_to(thr, temp_result, c)
    # Bootstrap without KeySwitch
    bootstrap(
        thr, u2, cloud_key.bootstrap_key, cloud_key.keyswitch_key, MU, temp_result,
        perf_params, no_keyswitch=True)

    # Add u1=u1+u2
    MuxConst = phase_to_t32(1, 8)
    lwe_noiseless_trivial_constant(thr, temp_result1, MuxConst)
    lwe_add_to(thr, temp_result1, u1)
    lwe_add_to(thr, temp_result1, u2)

    # Key switching
    lwe_keyswitch(thr, result, cloud_key.keyswitch_key, temp_result1)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy
from reikna import cluda

from .lwe import LweSampleArray
from .api_low_level import (
    NuFHEParameters, NuFHESecretKey, NuFHECloudKey, encrypt, decrypt, empty_ciphertext)
from .performance import PerformanceParameters
from . import gates
from .random_numbers import DeterministicRNG
from .computation_cache import clear_computation_cache
from .gates import get_shape, result_shape


def _get_api_object(api):
    """
    Returns a Reikna API object based on API name (``None``, ``'CUDA'`` or ``'OpenCL'``).
    """
    api_funcs = {
        None: cluda.any_api,
        'CUDA': cluda.cuda_api,
        'OpenCL': cluda.ocl_api,
        }
    if api not in api_funcs:
        raise ValueError("Uknonwn GPGPU API identifier: " + str(api))
    return api_funcs[api]()


def find_devices(api=None,
        include_devices=None, exclude_devices=None,
        include_platforms=None, exclude_platforms=None):
    """
    Returns a list of computation device identifiers for the given API and selection criteria.
    If there are several platforms with suitable devices, only the first one will be used
    (so if you need a specific platform, use the corresponding masks).

    :param api: the GPGPU backend to use, one of ``None``, ``"CUDA"`` and ``"OpenCL"``.
        If ``None`` is given, an arbitrary available backend will be chosen.
    :param include_devices: a list of strings; only devices with one of the strings
        present in the name will be included.
    :param exclude_devices: a list of strings; devices with one of the strings
        present in the name will be excluded.
    :param include_platforms: a list of strings; only platforms with one of the strings
        present in the name will be included.
    :param exclude_platforms: a list of strings; platforms with one of the strings
        present in the name will be excluded.
    :returns: a list of :py:class:`~nufhe.api_high_level.DeviceID` objects.
    """

    api_obj = _get_api_object(api)

    platforms_and_devices = cluda.find_devices(
        api_obj,
        include_devices=include_devices,
        exclude_devices=exclude_devices,
        include_platforms=include_platforms,
        exclude_platforms=exclude_platforms)

    if len(platforms_and_devices) == 0:
        raise ValueError("No devices found satisfying the given search criteria")

    platform_ids = list(platforms_and_devices)

    return [
        DeviceID(api_obj.get_id(), platform_ids[0], device_id)
        for device_id in platforms_and_devices[platform_ids[0]]]


class DeviceID:
    """
    An identifier of a computation device suitable to run NuFHE.
    Obtained from :py:func:`~nufhe.find_devices`.
    Can be passed to another thread/process and used to create a :py:class:`~nufhe.Context` object.

    .. py:attribute:: api_name

        The name of the API (``"CUDA"`` or ``"OpenCL"``).

    .. py:attribute:: platform_name

        The name of the platform.

    .. py:attribute:: device_name

        The name of the device.
    """

    def __init__(self, api_id, platform_id, device_id):
        self.api_id = api_id
        self.platform_id = platform_id
        self.device_id = device_id

        # We do not save the actual device_obj because this object
        # can be passed to another thread/process.
        api, platform, device = self._get_objects()
        self.platform_name = platform.name
        self.device_name = device.name
        self.api_name = "CUDA" if api.get_id() == cluda.cuda_id() else "OpenCL"

    def _get_objects(self):
        api_obj = cluda.get_api(self.api_id)
        platform = api_obj.get_platforms()[self.platform_id]
        return api_obj, platform, platform.get_devices()[self.device_id]

    def get_api_and_device(self):
        api_obj, _, device = self._get_objects()
        return api_obj, device

    def __str__(self):
        return "DeviceID({api}, {platform}, {device})".format(
            api=self.api_name, platform=self.platform_name, device=self.device_name)


class Context:
    """
    An object encapuslating an execution environment on a GPU.

    If ``thread`` is given, it will be used to create the context;
    otherwise, if ``device_id`` is given, it will be used;
    if none of the above is given, the first device satisfying the given criteria will be used.

    :param rng: a random number generator which will be used wherever randomness is required.
        Can be an instance of one of the :ref:`random-number-generators`
        (:py:class:`DeterministicRNG` by default).
    :param thread: a Reikna ``Thread`` object to use internally for the context.
    :param device_id: a GPGPU device (and API) to use for the context.
    :param interactive: if ``True``, an interactive dialogue will be shown
        allowing one to choose the GPGPU device to use.
        If ``False``, the first device satisfying the filters (see below) will be chosen.
    :param api:
    :param include_devices:
    :param exclude_devices:
    :param include_platforms:
    :param exclude_platforms: see :py:func:`~nufhe.find_devices`.
    """

    def __init__(
            self, rng=None,
            thread=None,
            device_id: DeviceID=None,
            api=None, interactive=False,
            include_devices=None, exclude_devices=None,
            include_platforms=None, exclude_platforms=None):

        if rng is None:
            rng = DeterministicRNG()

        if thread is not None:
            # Use the given Thread object
            pass
        elif device_id is not None:
            api_obj, device = device_id.get_api_and_device()
            thread = api_obj.Thread(device)
        else:
            api_obj = _get_api_object(api)
            thread = api_obj.Thread.create(
                interactive=interactive,
                device_filters=dict(
                    include_devices=include_devices,
                    exclude_devices=exclude_devices,
                    include_platforms=include_platforms,
                    exclude_platforms=exclude_platforms))

        self.rng = rng
        self.thread = thread

    def __del__(self):
        # Comnputation cache retains some Thread-related objects,
        # so it needs to be cleared before Thread is destroyed.
        # This helps avoid CUDA errors in multi-thread usage.
        if hasattr(self, 'thread'):
            clear_computation_cache(self.thread)

    def make_secret_key(self, **params):
        """
        Creates a secret key, with ``params`` used to
        initialize a :py:class:`NuFHEParameters` object.

        The low-level analogue: :py:meth:`NuFHESecretKey.from_rng`.

        :returns: a :py:class:`NuFHESecretKey` object.
        """
        nufhe_params = NuFHEParameters(**params)
        return NuFHESecretKey.from_rng(self.thread, nufhe_params, self.rng)

    def make_cloud_key(self, secret_key: NuFHESecretKey):
        """
        Creates a cloud key matching the given secret key.

        The low-level analogue: :py:meth:`NuFHECloudKey.from_rng`.

        :returns: a :py:class:`NuFHECloudKey` object.
        """
        return NuFHECloudKey.from_rng(self.thread, secret_key.params, self.rng, secret_key)

    def make_key_pair(self, **params):
        """
        Creates a pair of a secret key and a matching cloud key.

        The low-level analogue: :py:func:`make_key_pair`.

        :returns: a tuple of a :py:class:`NuFHESecretKey` and a :py:class:`NuFHECloudKey` objects.
        """
        secret_key = self.make_secret_key(**params)
        cloud_key = self.make_cloud_key(secret_key)
        return secret_key, cloud_key

    def encrypt(self, secret_key: NuFHESecretKey, message):
        """
        Encrypts a message (a list or a ``numpy`` array treated as an array of booleans).

        The low-level analogue: :py:func:`encrypt`.

        :returns: an :py:class:`LweSampleArray` object with the same `shape` as the given array.
        """
        return encrypt(self.thread, self.rng, secret_key, message)

    def decrypt(self, secret_key: NuFHESecretKey, ciphertext: LweSampleArray):
        """
        Decrypts a message.

        The low-level analogue: :py:func:`decrypt`.

        :returns: a ``numpy.ndarray`` object of the type ``numpy.bool``
            and the same `shape` as ``ciphertext``.
        """
        return decrypt(self.thread, secret_key, ciphertext)

    def make_virtual_machine(
            self, cloud_key: NuFHECloudKey, perf_params: PerformanceParameters=None):
        """
        Creates an FHE "virtual machine" which can execute logical gates using the given cloud key.
        Optionally, one can pass a :py:class:`PerformanceParameters` object which will be
        specialized for the GPU device of the context and used in all the gate calls.

        :returns: a :py:class:`~nufhe.api_high_level.VirtualMachine` object.
        """
        return VirtualMachine(self.thread, cloud_key, perf_params=perf_params)

    def load_ciphertext(self, file_or_bytestring):
        """
        Load a ciphertext (a :py:class:`LweSampleArray` object) serialized with
        :py:meth:`LweSampleArray.dump` or :py:meth:`LweSampleArray.dumps`
        into the context memory space.

        The low-level analogues: :py:meth:`LweSampleArray.load` and :py:meth:`LweSampleArray.loads`.

        :returns: an :py:class:`LweSampleArray` object
        """
        if isinstance(file_or_bytestring, bytes):
            return LweSampleArray.loads(file_or_bytestring, self.thread)
        else:
            return LweSampleArray.load(file_or_bytestring, self.thread)

    def load_secret_key(self, file_or_bytestring):
        """
        Load a secret key (a :py:class:`NuFHESecretKey` object) serialized with
        :py:meth:`NuFHESecretKey.dump` or :py:meth:`NuFHESecretKey.dumps`
        into the context memory space.

        The low-level analogues: :py:meth:`NuFHESecretKey.load` and :py:meth:`NuFHESecretKey.loads`.

        :returns: a :py:class:`NuFHESecretKey` object
        """
        if isinstance(file_or_bytestring, bytes):
            return NuFHESecretKey.loads(file_or_bytestring, self.thread)
        else:
            return NuFHESecretKey.load(file_or_bytestring, self.thread)

    def load_cloud_key(self, file_or_bytestring):
        """
        Load a secret key (a :py:class:`NuFHECloudKey` object) serialized with
        :py:meth:`NuFHECloudKey.dump` or :py:meth:`NuFHECloudKey.dumps`
        into the context memory space.

        The low-level analogues: :py:meth:`NuFHECloudKey.load` and :py:meth:`NuFHECloudKey.loads`.

        :returns: a :py:class:`NuFHECloudKey` object
        """
        if isinstance(file_or_bytestring, bytes):
            return NuFHECloudKey.loads(file_or_bytestring, self.thread)
        else:
            return NuFHECloudKey.load(file_or_bytestring, self.thread)


class VirtualMachine:
    """
    A fully encrypted virtual machine capable of executing gates on ciphertexts
    (:py:class:`~nufhe.LweSampleArray` objects) using an encapsulated cloud key.

    .. method:: gate_<operator>(*args, dest: LweSampleArray=None)

        Calls one of :ref:`logical-gates`, using the context, the cloud key,
        and the performance parameters of the virtual machine.

        If ``dest`` is ``None``, creates a new ciphertext and uses it
        to store the output of the gate;
        otherwise ``dest`` is used for that purpose.

        :returns: an :py:class:`~nufhe.LweSampleArray` object
            with the result of the gate application.
    """

    def __init__(self, thread, cloud_key: NuFHECloudKey, perf_params: PerformanceParameters=None):
        "__init__()"
        if perf_params is None:
            perf_params = PerformanceParameters(cloud_key.params)

        perf_params = perf_params.for_device(thread.device_params)

        self.thread = thread
        self.params = cloud_key.params
        self.cloud_key = cloud_key
        self.perf_params = perf_params

    def empty_ciphertext(self, shape):
        """
        Returns an unitialized ciphertext (an :py:class:`~nufhe.LweSampleArray` object).

        The low-level analogue: :py:func:`empty_ciphertext`.
        """
        return empty_ciphertext(self.thread, self.params, shape)

    def load_ciphertext(self, file):
        """
        Load a ciphertext (a :py:class:`~nufhe.LweSampleArray` object) serialized with
        :py:meth:`LweSampleArray.dump <nufhe.LweSampleArray.dump>` into the context memory space.

        The low-level analogue: :py:meth:`LweSampleArray.load <nufhe.LweSampleArray.load>`.

        :returns: an :py:class:`~nufhe.LweSampleArray` object
        """
        return LweSampleArray.load(file, self.thread)

    def _gate(self, name, *args, dest: LweSampleArray=None):
        if dest is None:
            shapes = [get_shape(arg) for arg in args]
            dest = self.empty_ciphertext(result_shape(*shapes))
        gate_func = getattr(gates, name)
        gate_func(self.thread, self.cloud_key, dest, *args, perf_params=self.perf_params)
        return dest

    def __getattr__(self, name):
        if name.startswith('gate_'):
            return lambda *args, **kwds: self._gate(name, *args, **kwds)
        else:
            raise AttributeError(name)



"""
This example demonstrates how to build a simple multi-threaded application,
where each thread uses a separate GPU to apply an FHE gate to a part of a ciphertext.

In order to run it you need two GPGPU devices available, both of them supporting
the transform type you are planning to use (because they will have to use the same cloud key).

This example uses multi-threading for simplicity, but multi-processing
(using either the Python module or MPI) is also possible.
Note that CUDA may have some restrictions, for example, you cannot have a CUDA
context both in the main process and the child processes when using Python multithreading.
"""

from threading import Thread
from queue import Queue

import random
import nufhe


class MyThread:
    """
    A simple wrapper that allows one to receive the value
    returned from the thread worker transparently.
    """

    def __init__(self, target, args=()):
        self.return_queue = Queue()
        self.target = target
        self.thread = Thread(target=self._target_wrapper, args=args)

    def _target_wrapper(self, *args):
        res = self.target(*args)
        self.return_queue.put(res)

    def start(self):
        self.thread.start()
        return self

    def join(self):
        ret_val = self.return_queue.get()
        self.thread.join()
        return ret_val


def worker(device_id, cloud_key_cpu, ciphertext1_cpu, ciphertext2_cpu):
    """
    The thread worker function.
    Runs a NAND gate over two provided ciphertexts and returns the serialized result.
    """
    print("Running a thread with", device_id)

    ctx = nufhe.Context(device_id=device_id)

    cloud_key = ctx.load_cloud_key(cloud_key_cpu)
    ciphertext1 = ctx.load_ciphertext(ciphertext1_cpu)
    ciphertext2 = ctx.load_ciphertext(ciphertext2_cpu)

    vm = ctx.make_virtual_machine(cloud_key)
    result = vm.gate_nand(ciphertext1, ciphertext2)
    result_cpu = result.dumps()

    print("Done")

    return result_cpu


if __name__ == '__main__':

    # This part is identical to the `gate_nand` example:
    # create the key pair and encrypt some data.

    size = 32
    bits1 = [random.choice([False, True]) for i in range(size)]
    bits2 = [random.choice([False, True]) for i in range(size)]
    reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

    ctx = nufhe.Context()
    secret_key, cloud_key = ctx.make_key_pair()

    ciphertext1 = ctx.encrypt(secret_key, bits1)
    ciphertext2 = ctx.encrypt(secret_key, bits2)

    # Serialize the cloud key to pass it to child threads.

    ck = cloud_key.dumps()

    # Split ciphertexts into two parts each and serialize them.

    ct1_part1 = ciphertext1[:size//2].dumps()
    ct1_part2 = ciphertext1[size//2:].dumps()
    ct2_part1 = ciphertext2[:size//2].dumps()
    ct2_part2 = ciphertext2[size//2:].dumps()

    # Start two threads each applying NAND gate to their respective parts of the ciphertext.

    devices = nufhe.find_devices()

    assert len(devices) >= 2

    t1 = MyThread(target=worker, args=(devices[0], ck, ct1_part1, ct2_part1)).start()
    t2 = MyThread(target=worker, args=(devices[1], ck, ct1_part2, ct2_part2)).start()
    result_part1 = t1.join()
    result_part2 = t2.join()

    result_part1 = ctx.load_ciphertext(result_part1)
    result_part2 = ctx.load_ciphertext(result_part2)

    # Decrypt the results, join and test against the reference

    r1 = ctx.decrypt(secret_key, result_part1)
    r2 = ctx.decrypt(secret_key, result_part2)

    assert r1.tolist() + r2.tolist() == reference



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.cluda import Module

from .numeric_functions_gpu import Torus32ToPhase
from .numeric_functions_gpu import Torus32, Int32, ErrorFloat # for re-export
from .computation_cache import get_computation


# Approximate the phase to the nearest message possible in the message space.
# The constant `mspace_size` indicates on which message space we are working
# (how many messages possible).
def phase_to_t32(phase: int, mspace_size: int):
    return Torus32((phase % mspace_size) * (2**32 // mspace_size))


def t32_to_phase(thr, result, messages, mspace_size: int):
    comp = get_computation(thr, Torus32ToPhase, messages.shape, mspace_size)
    comp(result, messages)


def double_to_t32(d: float):
    return ((d - numpy.trunc(d)) * 2**32).astype(Torus32)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import itertools
import numpy
import pytest

from reikna.cluda import ocl_id, cuda_id

from nufhe.transform import fft512, ntt1024, Transform
import nufhe.transform.fft as tr_fft
import nufhe.transform.ntt as tr_ntt
from nufhe.polynomial_transform import max_supported_transforms_per_block, transform_supported

from utils import get_test_array


@pytest.mark.parametrize('inverse', [False, True], ids=['forward', 'inverse'])
@pytest.mark.parametrize('i32_conversion', [False, True], ids=['no_conversion', 'poly_conversion'])
@pytest.mark.parametrize('constant_memory', [False, True], ids=['global_mem', 'constant_mem'])
def test_transform_correctness(thread, transform_type, inverse, i32_conversion, constant_memory):

    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()

    batch_shape = (128,)

    if transform_type == 'FFT':
        transform = fft512(use_constant_memory=constant_memory)
        transform_ref = tr_fft.fft_transform_ref
    else:
        transform = ntt1024(use_constant_memory=constant_memory)
        transform_ref = tr_ntt.ntt_transform_ref

    comp = Transform(
        transform, batch_shape,
        inverse=inverse, i32_conversion=i32_conversion, transforms_per_block=1,
        ).compile(thread)

    a = get_test_array(comp.parameter.input.shape, comp.parameter.input.dtype)

    a_dev = thread.to_device(a)
    res_dev = thread.empty_like(comp.parameter.output)

    comp(res_dev, a_dev)
    res_test = res_dev.get()

    res_ref = transform_ref(a, inverse=inverse, i32_conversion=i32_conversion)

    if numpy.issubdtype(res_dev.dtype, numpy.integer):
        assert (res_test == res_ref).all()
    else:
        assert numpy.allclose(res_test, res_ref)


def poly_mul_ref(p1, p2):
    N = p1.shape[-1]

    result = numpy.empty_like(p1)

    for i in range(N):
        result[:,i] = (p1[:,:i+1] * p2[:,i::-1]).sum(1) - (p1[:,i+1:] * p2[:,:i:-1]).sum(1)

    return result


def test_polynomial_multiplication(thread, transform_type):
    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()

    batch_shape = (10,)

    if transform_type == 'FFT':
        transform = fft512()
        transform_ref = tr_fft.fft_transform_ref
        tr_mul_ref = tr_fft.fft_transformed_mul_ref
    else:
        transform = ntt1024()
        transform_ref = tr_ntt.ntt_transform_ref
        tr_mul_ref = tr_ntt.ntt_transformed_mul_ref

    tr_forward = Transform(
        transform, batch_shape,
        inverse=False, i32_conversion=True, transforms_per_block=1,
        ).compile(thread)
    tr_inverse = Transform(
        transform, batch_shape,
        inverse=True, i32_conversion=True, transforms_per_block=1,
        ).compile(thread)

    a = numpy.random.randint(-2**31, 2**31, size=batch_shape + (1024,), dtype=numpy.int32)
    b = numpy.random.randint(-1000, 1000, size=batch_shape + (1024,), dtype=numpy.int32)

    a_dev = thread.to_device(a)
    b_dev = thread.to_device(b)
    a_tr_dev = thread.empty_like(tr_forward.parameter.output)
    b_tr_dev = thread.empty_like(tr_forward.parameter.output)
    res_dev = thread.empty_like(tr_inverse.parameter.output)

    tr_forward(a_tr_dev, a_dev)
    tr_forward(b_tr_dev, b_dev)
    res_tr = tr_mul_ref(a_tr_dev.get(), b_tr_dev.get())
    res_tr_dev = thread.to_device(res_tr)
    tr_inverse(res_dev, res_tr_dev)

    res_test = res_dev.get()
    res_ref = poly_mul_ref(a, b)

    assert (res_test == res_ref).all()


def get_times(thread, comp, out_arr, in_arr, attempts=10):
    # Test performance
    times = []

    for j in range(attempts):
        thread.synchronize()
        t1 = time.time()
        comp(out_arr, in_arr)
        thread.synchronize()
        t2 = time.time()
        times.append(t2 - t1)

    times = numpy.array(times)

    return times, "min: {min:.4f}, mean: {mean:.4f}, std: {std:.4f}".format(
        min=times.min(), mean=times.mean(), std=times.std())


@pytest.mark.perf
@pytest.mark.parametrize('transforms_per_block', [1, 2, 4])
@pytest.mark.parametrize('constant_memory', [False, True], ids=['global_mem', 'constant_mem'])
def test_ntt_performance(thread, transforms_per_block, constant_memory, heavy_performance_load):

    if not transform_supported(thread.device_params, 'NTT'):
        pytest.skip()

    if transforms_per_block > max_supported_transforms_per_block(thread.device_params, 'NTT'):
        pytest.skip()

    is_cuda = thread.api.get_id() == cuda_id()

    methods = list(itertools.product(
        ['cuda_asm', 'c'], # base method
        ['cuda_asm', 'c_from_asm', 'c'], # mul method
        ['cuda_asm', 'c_from_asm', 'c'] # lsh method
        ))

    if not is_cuda:
        # filter out all usage of CUDA asm if we're on OpenCL
        methods = [ms for ms in methods if 'cuda_asm' not in ms]

    batch_shape = (2**14,)
    a = get_test_array(batch_shape + (1024,), "ff_number")

    kernel_repetitions = 100 if heavy_performance_load else 5

    a_dev = thread.to_device(a)
    res_dev = thread.empty_like(a_dev)

    # TODO: compute a reference NTT when it's fast enough on CPU
    #res_ref = tr_ntt.ntt_transform_ref(a)

    print()
    min_times = []
    for base_method, mul_method, lsh_method in methods:

        transform = ntt1024(
            base_method=base_method, mul_method=mul_method, lsh_method=lsh_method,
            use_constant_memory=constant_memory)

        ntt_comp = Transform(
            transform, batch_shape, transforms_per_block=transforms_per_block,
            ).compile(thread)
        ntt_comp_repeated = Transform(
            transform, batch_shape, transforms_per_block=transforms_per_block,
            kernel_repetitions=kernel_repetitions).compile(thread)

        # TODO: compute a reference NTT when it's fast enough on CPU
        # Quick check of correctness
        #ntt_comp(res_dev, a_dev)
        #res_test = res_dev.get()
        #assert (res_test == res_ref).all()

        # Test performance
        times, times_str = get_times(thread, ntt_comp_repeated, res_dev, a_dev)
        print("  base: {bm}, mul: {mm}, lsh: {lm}".format(
            bm=base_method, mm=mul_method, lm=lsh_method))
        print("  {backend}, {trnum} per block, test --- {times}".format(
            times=times_str,
            backend='cuda' if is_cuda else 'ocl ',
            trnum=transforms_per_block))

        min_times.append((times.min(), base_method, mul_method, lsh_method))

    best = min(min_times, key=lambda t: t[0])
    time_best, base_method, mul_method, lsh_method = best
    print("Best time: {tb:.4f} for [base: {bm}, mul: {mm}, lsh: {lm}]".format(
        tb=time_best, bm=base_method, mm=mul_method, lm=lsh_method
        ))


@pytest.mark.perf
@pytest.mark.parametrize('transforms_per_block', [1, 2, 3, 4])
@pytest.mark.parametrize('constant_memory', [False, True], ids=['global_mem', 'constant_mem'])
def test_fft_performance(thread, transforms_per_block, constant_memory, heavy_performance_load):

    if not transform_supported(thread.device_params, 'FFT'):
        pytest.skip()

    if transforms_per_block > max_supported_transforms_per_block(thread.device_params, 'FFT'):
        pytest.skip()

    is_cuda = thread.api.get_id() == cuda_id()

    batch_shape = (2**14,)
    a = get_test_array(batch_shape + (512,), numpy.complex128)

    kernel_repetitions = 100 if heavy_performance_load else 5

    a_dev = thread.to_device(a)
    res_dev = thread.empty_like(a_dev)

    res_ref = tr_fft.fft_transform_ref(a)

    transform = fft512(use_constant_memory=constant_memory)

    fft_comp = Transform(
        transform, batch_shape, transforms_per_block=transforms_per_block,
        ).compile(thread)
    fft_comp_repeated = Transform(
        transform, batch_shape, transforms_per_block=transforms_per_block,
        kernel_repetitions=kernel_repetitions).compile(thread)

    # Quick check of correctness
    fft_comp(res_dev, a_dev)
    res_test = res_dev.get()
    assert numpy.allclose(res_test, res_ref)

    # Test performance
    times, times_str = get_times(thread, fft_comp_repeated, res_dev, a_dev)
    print("\n{backend}, {trnum} per block, test --- {times}".format(
        times=times_str,
        backend='cuda' if is_cuda else 'ocl ',
        trnum=transforms_per_block))



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Torus Gentry, Sahai & Waters bootstrapping functions.
"""

import pickle

import numpy

from .numeric_functions import Torus32
from .tlwe import (
    TLweParams,
    TLweKey,
    TLweSampleArray,
    TransformedTLweSampleArray,
    tlwe_transform_samples,
    tlwe_encrypt_zero,
    )
from .tgsw_gpu import (
    TGswAddMessage,
    TGswTransformedExternalMul,
    )
from .computation_cache import get_computation
from .performance import PerformanceParametersForDevice


class TGswParams:

    def __init__(self, tlwe_params: TLweParams, decomp_length: int, bs_log2_base: int):

        # 1/(base^(i+1)) as a Torus32
        decomp_range = numpy.arange(1, decomp_length + 1)
        self.base_powers = (2**(32 - decomp_range * bs_log2_base)).astype(Torus32)

        # offset = base/2 * Sum{j=1..decomp_length} 2^(32 - j * bs_log2_base)
        self.offset = (
            self.base_powers.astype(numpy.int64).sum() * (2**bs_log2_base // 2)).astype(Torus32)

        self.decomp_length = decomp_length
        self.bs_log2_base = bs_log2_base
        self.tlwe_params = tlwe_params # Params of each row

    def __eq__(self, other: 'TGswParams'):
        return (
            self.__class__ == other.__class__
            and self.decomp_length == other.decomp_length
            and self.bs_log2_base == other.bs_log2_base
            and self.tlwe_params == other.tlwe_params)

    def __hash__(self):
        return hash((self.__class__, self.decomp_length, self.bs_log2_base, self.tlwe_params))


class TGswKey:

    def __init__(self, params: TGswParams, tlwe_key: TLweKey):
        self.params = params
        self.tlwe_key = tlwe_key

    @classmethod
    def from_rng(cls, thr, params: TGswParams, rng):
        return cls(params, TLweKey.from_rng(thr, params.tlwe_params, rng))


class TGswSampleArray:

    def __init__(self, params: TGswParams, samples: TLweSampleArray):
        self.mask_size = params.tlwe_params.mask_size
        self.decomp_length = params.decomp_length
        self.samples = samples
        self.params = params
        self.shape = samples.shape[:-2]

    @classmethod
    def empty(cls, thr, params: TGswParams, shape):
        mask_size = params.tlwe_params.mask_size
        decomp_length = params.decomp_length
        samples = TLweSampleArray.empty(
            thr, params.tlwe_params, shape + (mask_size + 1, decomp_length))
        return cls(params, samples)


class TransformedTGswSampleArray:

    def __init__(self, params: TGswParams, samples: TransformedTLweSampleArray):
        self.mask_size = params.tlwe_params.mask_size
        self.decomp_length = params.decomp_length
        self.samples = samples
        self.params = params
        self.shape = samples.shape[:-2]

    @classmethod
    def empty(cls, thr, params: TGswParams, shape):
        mask_size = params.tlwe_params.mask_size
        decomp_length = params.decomp_length
        samples = TransformedTLweSampleArray.empty(
            thr, params.tlwe_params, shape + (mask_size + 1, decomp_length))
        return cls(params, samples)

    def dump(self, file_obj):
        pickle.dump(self.params, file_obj)
        self.samples.dump(file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        params = pickle.load(file_obj)
        samples = TransformedTLweSampleArray.load(file_obj, thr)
        return cls(params, samples)

    def __eq__(self, other: 'TransformedTGswSampleArray'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and self.samples == other.samples)


# For all the kpl TLWE samples composing the TGSW sample
# It computes the inverse FFT of the coefficients of the TLWE sample
def tgsw_transform_samples(
        thr, result: TransformedTGswSampleArray, source: TGswSampleArray,
        perf_params: PerformanceParametersForDevice):
    tlwe_transform_samples(thr, result.samples, source.samples, perf_params)


# result += message * H
def tgsw_add_message(thr, result: TGswSampleArray, messages):
    comp = get_computation(thr, TGswAddMessage, result.params, result.shape)
    comp(result.samples.a.coeffs, messages)


# Result = tGsw(0)
def tgsw_encrypt_zero(
        thr, rng, result: TGswSampleArray, noise: float, key: TGswKey,
        perf_params: PerformanceParametersForDevice):
    tlwe_encrypt_zero(thr, rng, result.samples, noise, key.tlwe_key, perf_params)


# encrypts a constant message
def tgsw_encrypt_int(
        thr, rng, result: TGswSampleArray, messages, noise: float, key: TGswKey,
        perf_params: PerformanceParametersForDevice):

    # TYPING: messages::Array{Int32, 1}
    tgsw_encrypt_zero(thr, rng, result, noise, key, perf_params)
    tgsw_add_message(thr, result, messages)


# external product: accum = gsw (*) accum
def tgsw_transformed_external_mul(
        thr, result: TLweSampleArray, bootstrap_key: TransformedTGswSampleArray, bk_row_idx: int,
        perf_params: PerformanceParametersForDevice):
    assert len(bootstrap_key.shape) == 1
    comp = get_computation(
        thr, TGswTransformedExternalMul,
        bootstrap_key.params, result.shape, bootstrap_key.shape[0], perf_params)
    comp(result.a.coeffs, bootstrap_key.samples.a.coeffs, bk_row_idx)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna import helpers

from .numeric_functions import Torus32
from .polynomial_transform import get_transform


def tgsw_polynomial_decomp_trf_reference(params: 'TGswParams', shape):

    tlwe_params = params.tlwe_params
    decomp_length = params.decomp_length
    mask_size = tlwe_params.mask_size
    polynomial_degree = tlwe_params.polynomial_degree

    bs_log2_base = params.bs_log2_base
    offset = params.offset
    base = 2**bs_log2_base
    maskMod = base - 1
    halfBg = 2**(bs_log2_base - 1)

    def _kernel(result, sample):

        decomp_shift = lambda p: 32 - p * bs_log2_base

        ps = numpy.arange(1, decomp_length + 1).reshape((1,) * len(shape) + (1, decomp_length, 1))
        sample_coefs = sample.reshape(shape + (mask_size + 1, 1, polynomial_degree))

        # do the decomposition
        numpy.copyto(result, (((sample_coefs + offset) >> decomp_shift(ps)) & maskMod) - halfBg)

    return _kernel


def tlwe_transformed_add_mul_to_trf_reference(
        params: 'TGswParams', shape, bk_len: int, perf_params):

    tlwe_params = params.tlwe_params
    decomp_length = params.decomp_length
    mask_size = tlwe_params.mask_size
    polynomial_degree = tlwe_params.polynomial_degree

    transform = get_transform(tlwe_params.transform_type)
    tlength = transform.transformed_length(polynomial_degree)

    def _kernel(result, sample, bootstrap_key, bk_row_idx):
        batch_len = helpers.product(shape)
        sample_view = sample.reshape(batch_len, mask_size + 1, decomp_length, 1, tlength)
        result_view = result.reshape(batch_len, mask_size + 1, tlength)

        result.fill(0)
        for mask_idx in range(mask_size + 1):
            for decomp_idx in range(decomp_length):
                numpy.copyto(
                    result_view,
                    transform.transformed_space_add_ref(
                        result_view,
                        transform.transformed_space_mul_prepared_ref(
                            sample_view[:, mask_idx, decomp_idx, :, :],
                            bootstrap_key[bk_row_idx, mask_idx, decomp_idx, :, :])))

    return _kernel


def TGswTransformedExternalMulReference(params: 'TGswParams', shape, bk_len, perf_params):

    tlwe_params = params.tlwe_params
    decomp_length = params.decomp_length
    mask_size = tlwe_params.mask_size
    polynomial_degree = tlwe_params.polynomial_degree

    transform = get_transform(params.tlwe_params.transform_type)
    tlength = transform.transformed_length(polynomial_degree)
    tdtype = transform.transformed_dtype()

    def _kernel(accum, bootstrap_key, bk_row_idx):

        sample = numpy.empty(shape + (mask_size + 1, decomp_length, polynomial_degree), Torus32)
        tr_accum = numpy.empty(shape + (mask_size + 1, tlength), tdtype)

        decomp = tgsw_polynomial_decomp_trf_reference(params, shape)
        add_mul = tlwe_transformed_add_mul_to_trf_reference(params, shape, bk_len, perf_params)

        decomp(sample, accum)
        tr_sample = transform.forward_transform_ref(sample)
        add_mul(tr_accum, tr_sample, bootstrap_key, bk_row_idx)
        numpy.copyto(accum, transform.inverse_transform_ref(tr_accum))

    return _kernel


def TGswAddMessageReference(params: 'TGswParams', shape):
    mask_size = params.tlwe_params.mask_size
    polynomial_degree = params.tlwe_params.polynomial_degree
    decomp_length = params.decomp_length
    base_powers = params.base_powers

    def _kernel(result_a, messages):
        result_a_view = result_a.reshape(
            helpers.product(messages.shape),
            mask_size + 1, decomp_length, mask_size + 1, polynomial_degree)
        messages = messages.flatten()

        inc = messages.reshape(messages.size, 1) * base_powers.reshape(1, decomp_length)

        for mask_idx in range(mask_size + 1):
            result_a_view[:, mask_idx, :, mask_idx, 0] += inc

    return _kernel



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy


class GaloisNumber:

    modulus = 2**64 - 2**32 + 1
    factors = [2, 3, 5, 17, 257, 65537] # prime factors of (modulus - 1)

    def __init__(self, val):
        self.val = int(val) % self.modulus

    def __add__(self, other: "GaloisNumber"):
        return GaloisNumber((self.val + other.val) % self.modulus)

    def __sub__(self, other: "GaloisNumber"):
        return GaloisNumber((self.val - other.val) % self.modulus)

    def __mul__(self, other: "GaloisNumber"):
        return GaloisNumber((self.val * other.val) % self.modulus)

    def __truediv__(self, other: "GaloisNumber"):
        return self * other.inverse()

    def __pow__(self, e: int):
        if e == 0:
            return GaloisNumber(1)
        elif e == 1:
            return self
        else:
            x = GaloisNumber(self.val)
            y = GaloisNumber(1)
            while e > 1:
                if e % 2 != 0:
                    y *= x
                x *= x
                e //= 2
            return x * y

    def inverse(self):
        return self**(self.modulus - 2)

    def __eq__(self, other):
        if isinstance(other, GaloisNumber):
            return self.val == other.val
        else:
            return self.val == other

    def __repr__(self):
        return "GaloisNumber(" + str(self.val) + ")"

    def __str__(self):
        return str(self.val) + "G"


gnum = numpy.vectorize(GaloisNumber)

def _gnum_to_i32(x):
    # Treats any value less than the half of the finite field modulus as a positive integer,
    # and anything greater that that as a negative integer -(modulus - x),
    # then truncate the result to the i32 range.
    med = x.modulus // 2
    val = x.val
    return numpy.int32(val & 0xffffffff) - (val > med)

gnum_to_i32 = numpy.vectorize(_gnum_to_i32)

gnum_to_u64 = numpy.vectorize(lambda x: numpy.uint64(x.val))


def find_generator(start=2):
    for w in range(start, GaloisNumber.modulus):
        w = GaloisNumber(w)
        for q in GaloisNumber.factors:
            if w**((GaloisNumber.modulus - 1) // q) == 1:
                break
        else:
            return w


def root_of_unity(N):
    """
    Returns a root of unity of order N (that is, x^N=1).
    Different roots produce different results.
    A standard approach is to use

        find_generator()**((GaloisNumber.modulus - 1) // N)

    To compare the results against the GPU version,
    we return the root the GPU implementation uses.
    """
    assert 2**32 % N == 0
    return GaloisNumber(0xa70dc47e4cbdf43f)**(2**32 // N)


def galois_fft_base(N, inverse):
    w = root_of_unity(N)
    if inverse:
        return GaloisNumber(1) / w
    else:
        return w


def galois_fft_inverse_coeff(N):
    return GaloisNumber(1) / GaloisNumber(N)


def ntt_naive(a, inverse):
    N = a.size
    w = galois_fft_base(N, inverse)

    result = gnum(numpy.zeros(N))
    for i in range(N):
        for j in range(N):
            result[i] += a[j] * w**(i * j)

    if inverse:
        return result * galois_fft_inverse_coeff(N)
    else:
        return result


def bitreverse(x, l):
    # Slow, but simple function. Needs some optimized algorithm in practice.
    b = ('{:0' + str(l) + 'b}').format(x)
    return int(b[::-1], 2)


def fft_generic(fft_base_func, fft_inverse_coeff_func, data, inverse):

    batch_shape, n = data.shape[:-1], data.shape[-1]
    data = data.reshape(numpy.prod((1,) + batch_shape), n).transpose().copy()

    logn = numpy.round(numpy.log2(n)).astype(numpy.int32)

    assert n >= 2
    assert n == 2**logn

    for i in range(n):
        j = bitreverse(i, logn)
        if j > i:
            data[[j, i]] = data[[i, j]]

    w = fft_base_func(n, inverse)

    for stage in range(logn):
        mmax = 2**stage
        istep = mmax * 2

        for m in range(mmax):
            tw = w**(m * 2**(logn - stage - 1))

            for i in range(m, n, istep):
                j = i + mmax

                temp = data[j] * tw
                data[j] = data[i] - temp
                data[i] += temp

    data = data.transpose().reshape(batch_shape + (n,))

    if inverse:
        return data * fft_inverse_coeff_func(n)
    else:
        return data


def ntt(data, inverse):
    return fft_generic(galois_fft_base, galois_fft_inverse_coeff, data, inverse)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from nufhe.numeric_functions import Torus32, Int32
from nufhe.numeric_functions_gpu import Torus32ToPhase
from nufhe.numeric_functions_cpu import Torus32ToPhaseReference

from utils import get_test_array, errors_allclose


def test_t32_to_phase(thread):

    mspace_size = 2048
    shape = (10, 20, 30)
    phase = get_test_array(shape, Torus32)
    result = numpy.empty(shape, Int32)

    phase_dev = thread.to_device(phase)
    result_dev = thread.empty_like(result)

    comp = Torus32ToPhase(shape, mspace_size).compile(thread)
    ref = Torus32ToPhaseReference(shape, mspace_size)

    comp(result_dev, phase_dev)
    result_test = result_dev.get()

    ref(result, phase)

    assert errors_allclose(result_test, result)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections import defaultdict


_computations = defaultdict(lambda: dict())


def clean_arg(arg):
    if hasattr(arg, 'dtype'):
        return ('Type', arg.dtype, arg.shape, arg.strides, arg.offset)
    else:
        return arg


def clear_computation_cache(thr):
    """
    Clear the cache of computation objects compiled for the given ``reikna`` thread ``thr``.
    This will help ensure a correct realease of the thread's resources when the other references
    to it go out of scope
    (which is especially important for multi-threading applications using CUDA).

    .. note::

        :py:class:`~nufhe.Context` objects call this function automatically on destruction.
    """
    if id(thr) in _computations:
        del _computations[id(thr)]


def get_computation(thr, cls, *args, **kwds):
    hashable_args = tuple(map(clean_arg, args))
    hashable_kwds = tuple((key, kwds[key]) for key in sorted(kwds))
    key = (id(cls), hashable_args, hashable_kwds)
    if key in _computations[id(thr)]:
        return _computations[id(thr)][key]
    else:
        comp = cls(*args, **kwds)
        compiled_comp = comp.compile(thr)
        _computations[id(thr)][key] = compiled_comp
        return compiled_comp



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from .numeric_functions import Torus32, double_to_t32


def vec_mul_mat(a, b):
    return (a * b).sum(-1, dtype=Torus32)


def lwe_encrypt_with_external_noise(
        ks_a, ks_b, ks_cv, messages, noises_a, noises_b, noise: float, key):

    # term h=0 as trivial encryption of 0
    ks_a[:,:,0,:] = 0
    ks_b[:,:,0] = 0
    ks_cv[:,:,0] = 0

    ks_a[:,:,1:,:] = noises_a
    ks_b[:,:,1:] = messages + noises_b + vec_mul_mat(noises_a, key)
    ks_cv[:,:,1:] = noise**2


def MakeLweKeyswitchKeyReference(
        input_size: int, output_size: int, decomp_length: int, log2_base: int, noise: float):

    base = 2**log2_base

    def _kernel(ks_a, ks_b, ks_cv, in_key, out_key, noises_a, noises_b):

        hs = numpy.arange(1, base).astype(Torus32)
        js = numpy.arange(decomp_length).astype(Torus32)

        r_key = in_key[:, None, None]
        r_hs = hs[None, None, :]
        r_js = js[None, :, None]

        messages = r_key * r_hs * (2**(32 - (r_js + 1) * log2_base))

        lwe_encrypt_with_external_noise(
            ks_a, ks_b, ks_cv, messages, noises_a, noises_b, noise, out_key)

    return _kernel


def LweKeyswitchReference(
        shape_info, input_size: int, output_size: int, decomp_length: int, log2_base: int):

    def _kernel(result_a, result_b, result_cv, ks_a, ks_b, ks_cv, source_a, source_b):

        base = 2**log2_base
        prec_offset = 2**(32 - (1 + log2_base * decomp_length)) # precision
        mask = base - 1

        js = numpy.arange(1, decomp_length + 1).reshape(
            (1,) * len(source_a.shape) + (decomp_length,))
        source_a = source_a.reshape(source_a.shape + (1,))
        aijs = (((source_a + prec_offset) >> (32 - js * log2_base)) & mask)

        # Starting from a noiseless trivial LWE:
        # a = 0, b = bi, current_variances = 0
        result_a.fill(0)
        numpy.copyto(result_b, source_b)
        result_cv.fill(0)

        for l in range(input_size):
            for j in range(decomp_length):
                x = aijs.take(l, axis=-2).take(j, axis=-1)
                lwe_sub_to(result_a, result_b, result_cv, ks_a[l,j,x], ks_b[l,j,x], ks_cv[l,j,x])

    return _kernel


def lwe_sub_to(result_a, result_b, result_cv, source_a, source_b, source_cv):
    result_a -= source_a
    result_b -= source_b
    result_cv += source_cv


def LweEncryptReference(shape, lwe_size: int, noise: float):

    def _kernel(result_a, result_b, result_cv, messages, key, noises_a, noises_b):
        numpy.copyto(result_b, noises_b + messages)
        numpy.copyto(result_a, noises_a)
        result_b += vec_mul_mat(result_a, key)
        result_cv.fill(noise**2)

    return _kernel


def LweDecryptReference(shape, lwe_size: int):

    def _kernel(result, lwe_a, lwe_b, key):
        numpy.copyto(result, lwe_b - vec_mul_mat(lwe_a, key))

    return _kernel


def LweLinearReference(result_shape_info, source_shape_info, add_result=False):

    def _kernel(result_a, result_b, result_cv, source_a, source_b, source_cv, p):
        p = Torus32(p)
        numpy.copyto(result_a, (result_a if add_result else 0) + p * source_a)
        numpy.copyto(result_b, (result_b if add_result else 0) + p * source_b)
        numpy.copyto(result_cv, (result_cv if add_result else 0) + p**2 * source_cv)

    return _kernel


def LweNoiselessTrivialReference(result_shape_info, source_shape):

    def _kernel(result_a, result_b, result_cv, mus):
        result_a.fill(0)
        numpy.copyto(result_b, mus)
        result_cv.fill(0)

    return _kernel


def LweNoiselessTrivialConstantReference(result_shape_info):

    def _kernel(result_a, result_b, result_cv, mu):
        result_a.fill(0)
        result_b.fill(mu)
        result_cv.fill(0)

    return _kernel



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple
from reikna.cluda import cuda_id


class PerformanceParameters:
    """
    Advanced performance settings for bootstrapping.

    For all the optional parameters below, if ``None`` is given,
    the library will attempt to select the best performing variant,
    given the available information.

    :param nufhe_params: a :py:class:`NuFHEParameters` object.
    :param ntt_base_method: ``'cuda_asm'`` or ``'c'``;
        An algorithm used in NTT for modulo addition, modulo subtraction, and modulus.
    :param ntt_mul_method: one of ``'cuda_asm'``, ``'c_from_asm'`` and ``'c'``;
        An algorithm used in NTT for modulo multiplication.
    :param ntt_lsh_method: one of ``'cuda_asm'``, ``'c_from_asm'`` and ``'c'``;
        An algorithm used in NTT for modulo bitshift.

    .. note::

        ``'cuda_asm'`` is only available for CUDA backend.
        When available, it is usually the fastest variant, or close to it.

    :param use_constant_memory_multi_iter: use constant GPU memory (as opposed to global memory)
        for precalculated coefficients in NTT/FFT in kernels where one of these transformations
        is executed multiple times per kernel call.
    :param use_constant_memory_single_iter: use constant GPU memory (as opposed to global memory)
        for precalculated coefficients in NTT/FFT in kernels where one of these transformations
        is executed once per kernel call.

    .. note::

        Using constant memory is usually beneficial on fast videocards
        if the transformation is executed many times per kernel call.

    :param transforms_per_block: a positive integer value, denoting how many separate transforms
        to execute in parallel on a single GPU multiprocessor (compute unit).

    .. note::

        On most videocards 1 to 4 transforms is supported for NTT, and 1 to 8 for FFT.
        More transforms does not necessarily mean better performance, since parallel threads
        on the same compute unit compete for resources.

        Since it is not trivial to determine the maximum in advance, if the requested number
        is greater than that, it will be dynamically reduced to the maximum possible value.

    :param single_kernel_bootstrap: if ``True``, execute bootstrap in a single kernel,
        instead of many separate kernel calls in a loop.

    .. note::

        Single kernel bootstrap is only supported for default FHE parameters,
        and needs the videocard to support a certain amount of parallel threads on a compute unit
        (256 for FFT, 512 for NTT).
        If available, it is usually significantly faster,
        partially due to lower kernel call overhead.

    :param low_end_device: if ``True``, the optimal values for low-end videocards will be picked.
        If ``None``, the decision will be made based on the number of compute units the device has.
    """

    __attributes__ = (
        'nufhe_params',
        'ntt_base_method',
        'ntt_mul_method',
        'ntt_lsh_method',
        'use_constant_memory_multi_iter',
        'use_constant_memory_single_iter',
        'transforms_per_block',
        'single_kernel_bootstrap',
        'low_end_device',
        )

    def __init__(
            self,
            nufhe_params, # TODO: type annotation here ('NuFHEParameters') triggers a Sphinx error.
            ntt_base_method=None,
            ntt_mul_method=None,
            ntt_lsh_method=None,
            use_constant_memory_multi_iter=None,
            use_constant_memory_single_iter=None,
            transforms_per_block=None,
            single_kernel_bootstrap=None,
            low_end_device=None):

        assert ntt_base_method in (None, 'cuda_asm', 'c')
        assert ntt_mul_method in (None, 'cuda_asm', 'c_from_asm', 'c')
        assert ntt_lsh_method in (None, 'cuda_asm', 'c_from_asm', 'c')

        self.nufhe_params = nufhe_params

        self.ntt_base_method = ntt_base_method
        self.ntt_mul_method = ntt_mul_method
        self.ntt_lsh_method = ntt_lsh_method
        self.use_constant_memory_multi_iter = use_constant_memory_multi_iter
        self.use_constant_memory_single_iter = use_constant_memory_single_iter
        self.transforms_per_block = transforms_per_block
        self.single_kernel_bootstrap = single_kernel_bootstrap
        self.low_end_device = low_end_device

    def for_device(self, device_params):
        """
        Specialize performance parameters for the given device
        (using a Reikna ``DeviceParams`` object).

        :returns: a :py:class:`~nufhe.performance.PerformanceParametersForDevice` object.
        """
        return PerformanceParametersForDevice(self, device_params)

    def __hash__(self):
        return hash((self.__class__,) + self.__attributes__)

    def __eq__(self, other):
        return all(getattr(self, attr) == getattr(other, attr) for attr in self.__attributes__)


class PerformanceParametersForDevice:

    __attributes__ = (
        'ntt_base_method',
        'ntt_mul_method',
        'ntt_lsh_method',
        'use_constant_memory_multi_iter',
        'use_constant_memory_single_iter',
        'transforms_per_block',
        'single_kernel_bootstrap',
        )

    def __init__(self, perf_params: PerformanceParameters, device_params):

        low_end_device = perf_params.low_end_device
        if low_end_device is None:
            # TODO: an arbitrary distinction, need to test on some devices close to it.
            low_end_device = device_params.compute_units < 20

        is_cuda = device_params.api_id == cuda_id()

        transform_type = perf_params.nufhe_params.tgsw_params.tlwe_params.transform_type
        mask_size = perf_params.nufhe_params.tgsw_params.tlwe_params.mask_size
        decomp_length = perf_params.nufhe_params.tgsw_params.decomp_length

        use_constant_memory_multi_iter = perf_params.use_constant_memory_multi_iter
        if use_constant_memory_multi_iter is None:
            use_constant_memory_multi_iter = not low_end_device

        use_constant_memory_single_iter = perf_params.use_constant_memory_single_iter
        if use_constant_memory_single_iter is None:
            use_constant_memory_single_iter = False

        # Avoiding circular reference
        from .polynomial_transform import max_supported_transforms_per_block
        max_supported_tpb = max_supported_transforms_per_block(device_params, transform_type)

        transforms_per_block = perf_params.transforms_per_block
        if transforms_per_block is None:
            if low_end_device:
                transforms_per_block = 1
            else:
                transforms_per_block = 4 if transform_type == 'NTT' else 2
            transforms_per_block = min(transforms_per_block, max_supported_tpb)
        else:
            if transforms_per_block > max_supported_tpb:
                raise ValueError(
                    "The chosen device does not support more than " + str(max_supported_tpb) +
                    " transforms per block")

        # Avoiding circular reference
        from .blind_rotate import single_kernel_bootstrap_supported

        single_kernel_bootstrap = perf_params.single_kernel_bootstrap
        skbs_supported = single_kernel_bootstrap_supported(perf_params.nufhe_params, device_params)

        if single_kernel_bootstrap is None:
            # If both encryption parameters and device capabilities allow it,
            # single kernel bootstrap is the optimal choice.
            single_kernel_bootstrap = not low_end_device and skbs_supported
        elif single_kernel_bootstrap and not skbs_supported:
            single_kernel_bootstrap_supported(
                perf_params.nufhe_params, device_params, raise_exception=True)

        ntt_base_method = perf_params.ntt_base_method
        ntt_mul_method = perf_params.ntt_mul_method
        ntt_lsh_method = perf_params.ntt_lsh_method

        if 'cuda_asm' in (ntt_base_method, ntt_mul_method, ntt_lsh_method) and not is_cuda:
            raise ValueError("'cuda_asm' is only supported for the CUDA backend")

        if low_end_device:
            skbs = single_kernel_bootstrap
            if ntt_base_method is None:
                ntt_base_method = ('c' if skbs else 'cuda_asm') if is_cuda else 'c'
            if ntt_mul_method is None:
                ntt_mul_method = ('c_from_asm' if skbs else 'cuda_asm') if is_cuda else 'c'
            if ntt_lsh_method is None:
                ntt_lsh_method = 'cuda_asm' if is_cuda else 'c'
        else:
            if ntt_base_method is None:
                ntt_base_method = 'cuda_asm' if is_cuda else 'c'
            if ntt_mul_method is None:
                ntt_mul_method = 'cuda_asm' if is_cuda else 'c'
            if ntt_lsh_method is None:
                ntt_lsh_method = 'cuda_asm' if is_cuda else 'c'

        self.ntt_base_method = ntt_base_method
        self.ntt_mul_method = ntt_mul_method
        self.ntt_lsh_method = ntt_lsh_method
        self.use_constant_memory_multi_iter = use_constant_memory_multi_iter
        self.use_constant_memory_single_iter = use_constant_memory_single_iter
        self.transforms_per_block = transforms_per_block
        self.single_kernel_bootstrap = single_kernel_bootstrap

    def __hash__(self):
        return hash((self.__class__,) + self.__attributes__)

    def __eq__(self, other):
        return all(getattr(self, attr) == getattr(other, attr) for attr in self.__attributes__)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import numpy

from nufhe.api_low_level import NuFHEParameters
from nufhe.lwe import LweSampleArrayShapeInfo, LweSampleArray, concatenate
from nufhe.lwe_gpu import (
    LweKeyswitch,
    MakeLweKeyswitchKey,
    LweEncrypt,
    LweDecrypt,
    LweLinear,
    LweNoiselessTrivial,
    LweNoiselessTrivialConstant,
    )
from nufhe.lwe_cpu import (
    LweKeyswitchReference,
    MakeLweKeyswitchKeyReference,
    LweEncryptReference,
    LweDecryptReference,
    LweLinearReference,
    LweNoiselessTrivialReference,
    LweNoiselessTrivialConstantReference,
    )
from nufhe.numeric_functions import Torus32, Int32, ErrorFloat, double_to_t32
import nufhe.random_numbers as rn

from utils import get_test_array, errors_allclose


def test_lwe_keyswitch(thread):

    batch_shape = (4, 5)

    params = NuFHEParameters()
    input_size = params.tgsw_params.tlwe_params.extracted_lweparams.size
    output_size = params.in_out_params.size
    decomp_length = params.ks_decomp_length
    log2_base = params.ks_log2_base
    base = 2**log2_base

    result_a = numpy.empty(batch_shape + (output_size,), Torus32)
    result_b = numpy.empty(batch_shape, Torus32)
    result_cv = numpy.empty(batch_shape, ErrorFloat)
    ks_a = get_test_array((input_size, decomp_length, base, output_size), Torus32, (-1000, 1000))
    ks_b = get_test_array((input_size, decomp_length, base), Torus32, (-1000, 1000))
    ks_cv = get_test_array((input_size, decomp_length, base), ErrorFloat, (-1, 1))

    # The base=0 slice of the keyswitch key is a "padding" - it's filled with zeroes.
    # The keyswitch function may rely on that.
    ks_a[:,:,0,:] = 0
    ks_b[:,:,0] = 0
    ks_cv[:,:,0] = 0

    source_a = get_test_array(batch_shape + (input_size,), Torus32)
    source_b = get_test_array(batch_shape, Torus32, (-1000, 1000))

    result_a_dev = thread.empty_like(result_a)
    result_b_dev = thread.empty_like(result_b)
    result_cv_dev = thread.empty_like(result_cv)
    ks_a_dev = thread.to_device(ks_a)
    ks_b_dev = thread.to_device(ks_b)
    ks_cv_dev = thread.to_device(ks_cv)
    source_a_dev = thread.to_device(source_a)
    source_b_dev = thread.to_device(source_b)

    shape_info = LweSampleArrayShapeInfo(result_a_dev, result_b_dev, result_cv_dev)
    test = LweKeyswitch(
        shape_info, input_size, output_size, decomp_length, log2_base).compile(thread)
    ref = LweKeyswitchReference(
        shape_info, input_size, output_size, decomp_length, log2_base)

    test(
        result_a_dev, result_b_dev, result_cv_dev,
        ks_a_dev, ks_b_dev, ks_cv_dev,
        source_a_dev, source_b_dev)
    result_a_test = result_a_dev.get()
    result_b_test = result_b_dev.get()
    result_cv_test = result_cv_dev.get()

    ref(result_a, result_b, result_cv, ks_a, ks_b, ks_cv, source_a, source_b)

    assert (result_a == result_a_test).all()
    assert (result_b == result_b_test).all()
    assert errors_allclose(result_cv, result_cv_test)


def test_make_lwe_keyswitch_key(thread):

    params = NuFHEParameters()
    input_size = params.tgsw_params.tlwe_params.extracted_lweparams.size
    output_size = params.in_out_params.size
    decomp_length = params.ks_decomp_length
    log2_base = params.ks_log2_base
    base = 2**log2_base
    noise = params.in_out_params.min_noise

    ks_a = numpy.empty((input_size, decomp_length, base, output_size), dtype=Torus32)
    ks_b = numpy.empty((input_size, decomp_length, base), dtype=Torus32)
    ks_cv = numpy.empty((input_size, decomp_length, base), dtype=ErrorFloat)

    in_key = get_test_array(input_size, Int32, (0, 2))
    out_key = get_test_array(output_size, Int32, (0, 2))
    noises_a = get_test_array((input_size, decomp_length, base - 1, output_size), Torus32)
    noises_b = double_to_t32(
        get_test_array((input_size, decomp_length, base - 1), ErrorFloat, (-noise, noise)))

    test = MakeLweKeyswitchKey(
        input_size, output_size, decomp_length, log2_base, noise).compile(thread)
    ref = MakeLweKeyswitchKeyReference(
        input_size, output_size, decomp_length, log2_base, noise)

    ks_a_dev = thread.empty_like(ks_a)
    ks_b_dev = thread.empty_like(ks_b)
    ks_cv_dev = thread.empty_like(ks_cv)
    in_key_dev = thread.to_device(in_key)
    out_key_dev = thread.to_device(out_key)
    noises_a_dev = thread.to_device(noises_a)
    noises_b_dev = thread.to_device(noises_b)

    test(ks_a_dev, ks_b_dev, ks_cv_dev, in_key_dev, out_key_dev, noises_a_dev, noises_b_dev)
    ref(ks_a, ks_b, ks_cv, in_key, out_key, noises_a, noises_b)

    ks_a_test = ks_a_dev.get()
    ks_b_test = ks_b_dev.get()
    ks_cv_test = ks_cv_dev.get()

    assert (ks_a_test == ks_a).all()
    assert (ks_b_test == ks_b).all()
    assert errors_allclose(ks_cv_test, ks_cv)


def test_lwe_encrypt(thread):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size
    noise = params.in_out_params.min_noise

    shape = (16, 20)
    result_a = numpy.empty(shape + (lwe_size,), Torus32)
    result_b = numpy.empty(shape, Torus32)
    result_cv = numpy.empty(shape, ErrorFloat)
    key = get_test_array(lwe_size, Int32, (0, 2))
    messages = get_test_array(shape, Torus32)
    noises_a = get_test_array(shape + (lwe_size,), Torus32)
    noises_b = get_test_array(shape, Torus32)

    test = LweEncrypt(shape, lwe_size, noise).compile(thread)
    ref = LweEncryptReference(shape, lwe_size, noise)

    result_a_dev = thread.empty_like(result_a)
    result_b_dev = thread.empty_like(result_b)
    result_cv_dev = thread.empty_like(result_cv)
    key_dev = thread.to_device(key)
    messages_dev = thread.to_device(messages)
    noises_a_dev = thread.to_device(noises_a)
    noises_b_dev = thread.to_device(noises_b)

    test(
        result_a_dev, result_b_dev, result_cv_dev,
        messages_dev, key_dev, noises_a_dev, noises_b_dev)
    ref(result_a, result_b, result_cv, messages, key, noises_a, noises_b)

    result_a_test = result_a_dev.get()
    result_b_test = result_b_dev.get()
    result_cv_test = result_cv_dev.get()

    assert (result_a_test == result_a).all()
    assert (result_b_test == result_b).all()
    assert errors_allclose(result_cv_test, result_cv)


def test_lwe_decrypt(thread):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    shape = (16, 20)
    result = numpy.empty(shape, Torus32)
    lwe_a = get_test_array(shape + (lwe_size,), Torus32)
    lwe_b = get_test_array(shape, Torus32)
    key = get_test_array(lwe_size, Int32, (0, 2))

    test = LweDecrypt(shape, lwe_size).compile(thread)
    ref = LweDecryptReference(shape, lwe_size)

    result_dev = thread.empty_like(result)
    lwe_a_dev = thread.to_device(lwe_a)
    lwe_b_dev = thread.to_device(lwe_b)
    key_dev = thread.to_device(key)

    test(result_dev, lwe_a_dev, lwe_b_dev, key_dev)
    ref(result, lwe_a, lwe_b, key)

    result_test = result_dev.get()

    assert (result_test == result).all()


@pytest.mark.parametrize('positive_coeff', [False, True], ids=['p<0', 'p>0'])
@pytest.mark.parametrize('add_result', [False, True], ids=['replace_result', 'update_result'])
def test_lwe_linear(thread, positive_coeff, add_result):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    shape = (10, 20)

    res_a = get_test_array(shape + (lwe_size,), Torus32)
    res_b = get_test_array(shape, Torus32)
    res_cv = get_test_array(shape, ErrorFloat, (-1, 1))

    src_a = get_test_array(shape + (lwe_size,), Torus32)
    src_b = get_test_array(shape, Torus32)
    src_cv = get_test_array(shape, ErrorFloat, (-1, 1))

    coeff = 1 if positive_coeff else -1

    shape_info = LweSampleArrayShapeInfo(src_a, src_b, src_cv)

    test = LweLinear(shape_info, shape_info, add_result=add_result).compile(thread)
    ref = LweLinearReference(shape_info, shape_info, add_result=add_result)

    res_a_dev = thread.to_device(res_a)
    res_b_dev = thread.to_device(res_b)
    res_cv_dev = thread.to_device(res_cv)
    src_a_dev = thread.to_device(src_a)
    src_b_dev = thread.to_device(src_b)
    src_cv_dev = thread.to_device(src_cv)
    thread.synchronize()

    test(res_a_dev, res_b_dev, res_cv_dev, src_a_dev, src_b_dev, src_cv_dev, coeff)
    ref(res_a, res_b, res_cv, src_a, src_b, src_cv, coeff)

    assert (res_a_dev.get() == res_a).all()
    assert (res_b_dev.get() == res_b).all()
    assert errors_allclose(res_cv_dev.get(), res_cv)


def test_lwe_linear_broadcast(thread):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    res_shape = (10, 20)
    src_shape = res_shape[1:]

    res_a = get_test_array(res_shape + (lwe_size,), Torus32)
    res_b = get_test_array(res_shape, Torus32)
    res_cv = get_test_array(res_shape, ErrorFloat, (-1, 1))

    src_a = get_test_array(src_shape + (lwe_size,), Torus32)
    src_b = get_test_array(src_shape, Torus32)
    src_cv = get_test_array(src_shape, ErrorFloat, (-1, 1))

    coeff = 1
    add_result = True

    res_shape_info = LweSampleArrayShapeInfo(res_a, res_b, res_cv)
    src_shape_info = LweSampleArrayShapeInfo(src_a, src_b, src_cv)

    test = LweLinear(res_shape_info, src_shape_info, add_result=add_result).compile(thread)
    ref = LweLinearReference(res_shape_info, src_shape_info, add_result=add_result)

    res_a_dev = thread.to_device(res_a)
    res_b_dev = thread.to_device(res_b)
    res_cv_dev = thread.to_device(res_cv)
    src_a_dev = thread.to_device(src_a)
    src_b_dev = thread.to_device(src_b)
    src_cv_dev = thread.to_device(src_cv)
    thread.synchronize()

    test(res_a_dev, res_b_dev, res_cv_dev, src_a_dev, src_b_dev, src_cv_dev, coeff)
    ref(res_a, res_b, res_cv, src_a, src_b, src_cv, coeff)

    assert (res_a_dev.get() == res_a).all()
    assert (res_b_dev.get() == res_b).all()
    assert errors_allclose(res_cv_dev.get(), res_cv)


def test_lwe_noiseless_trivial_constant(thread):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    shape = (10, 20)

    res_a = numpy.empty(shape + (lwe_size,), Torus32)
    res_b = numpy.empty(shape, Torus32)
    res_cv = numpy.empty(shape, ErrorFloat)
    mu = Torus32(-5)

    shape_info = LweSampleArrayShapeInfo(res_a, res_b, res_cv)

    test = LweNoiselessTrivialConstant(shape_info).compile(thread)
    ref = LweNoiselessTrivialConstantReference(shape_info)

    res_a_dev = thread.empty_like(res_a)
    res_b_dev = thread.empty_like(res_b)
    res_cv_dev = thread.empty_like(res_cv)

    test(res_a_dev, res_b_dev, res_cv_dev, mu)
    ref(res_a, res_b, res_cv, mu)

    assert (res_a_dev.get() == res_a).all()
    assert (res_b_dev.get() == res_b).all()
    assert errors_allclose(res_cv_dev.get(), res_cv)


def test_lwe_noiseless_trivial(thread):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    shape = (10, 20)

    res_a = numpy.empty(shape + (lwe_size,), Torus32)
    res_b = numpy.empty(shape, Torus32)
    res_cv = numpy.empty(shape, ErrorFloat)
    mus = get_test_array(shape, Torus32)

    shape_info = LweSampleArrayShapeInfo(res_a, res_b, res_cv)

    test = LweNoiselessTrivial(shape_info, shape).compile(thread)
    ref = LweNoiselessTrivialReference(shape_info, shape)

    res_a_dev = thread.empty_like(res_a)
    res_b_dev = thread.empty_like(res_b)
    res_cv_dev = thread.empty_like(res_cv)
    mus_dev = thread.to_device(mus)

    test(res_a_dev, res_b_dev, res_cv_dev, mus_dev)
    ref(res_a, res_b, res_cv, mus)

    assert (res_a_dev.get() == res_a).all()
    assert (res_b_dev.get() == res_b).all()
    assert errors_allclose(res_cv_dev.get(), res_cv)


@pytest.mark.parametrize('src_len', [0, 1])
def test_lwe_noiseless_trivial_broadcast(thread, src_len):

    params = NuFHEParameters()
    lwe_size = params.in_out_params.size

    res_shape = (10, 20)
    src_shape = res_shape[len(res_shape)-src_len:]

    res_a = numpy.empty(res_shape + (lwe_size,), Torus32)
    res_b = numpy.empty(res_shape, Torus32)
    res_cv = numpy.empty(res_shape, ErrorFloat)
    mus = get_test_array(src_shape, Torus32)

    shape_info = LweSampleArrayShapeInfo(res_a, res_b, res_cv)

    test = LweNoiselessTrivial(shape_info, src_shape).compile(thread)
    ref = LweNoiselessTrivialReference(shape_info, src_shape)

    res_a_dev = thread.empty_like(res_a)
    res_b_dev = thread.empty_like(res_b)
    res_cv_dev = thread.empty_like(res_cv)
    mus_dev = thread.to_device(mus)

    test(res_a_dev, res_b_dev, res_cv_dev, mus_dev)
    ref(res_a, res_b, res_cv, mus)

    assert (res_a_dev.get() == res_a).all()
    assert (res_b_dev.get() == res_b).all()
    assert errors_allclose(res_cv_dev.get(), res_cv)


def mock_ciphertext(thread, params, shape):
    ciphertext = LweSampleArray.empty(thread, params, shape)
    ciphertext.a = thread.to_device(get_test_array(ciphertext.a.shape, ciphertext.a.dtype))
    ciphertext.b = thread.to_device(get_test_array(ciphertext.b.shape, ciphertext.b.dtype))
    ciphertext.current_variances = thread.to_device(
        get_test_array(ciphertext.current_variances.shape, ciphertext.current_variances.dtype))
    return ciphertext


def test_lwe_copy(thread):

    params = NuFHEParameters()
    lwe_params = params.in_out_params

    shape = (3, 4, 5)

    ciphertext = mock_ciphertext(thread, lwe_params, shape)
    ciphertext_copy = ciphertext.copy()

    assert ciphertext == ciphertext_copy
    assert ciphertext.a is not ciphertext_copy.a
    assert ciphertext.b is not ciphertext_copy.b
    assert ciphertext.current_variances is not ciphertext_copy.current_variances


@pytest.mark.parametrize('shift', [7, -9, 0])
@pytest.mark.parametrize('axis', [0, 1, -1])
def test_lwe_roll(thread, shift, axis):

    params = NuFHEParameters()
    lwe_params = params.in_out_params

    shape = (3, 4, 5)

    ciphertext = mock_ciphertext(thread, lwe_params, shape)

    ciphertext_rolled = ciphertext.copy()
    ciphertext_rolled.roll(shift, axis=axis)

    src_a = ciphertext.a.get()
    src_b = ciphertext.b.get()
    src_cv = ciphertext.current_variances.get()

    res_a = ciphertext_rolled.a.get()
    res_b = ciphertext_rolled.b.get()
    res_cv = ciphertext_rolled.current_variances.get()

    roll_axis = axis % len(shape)

    assert (numpy.roll(src_a, shift, roll_axis) == res_a).all()
    assert (numpy.roll(src_b, shift, roll_axis) == res_b).all()
    assert numpy.allclose(numpy.roll(src_cv, shift, roll_axis), res_cv)


@pytest.mark.parametrize('axis', [0, 1])
@pytest.mark.parametrize('out_none', [False, True])
def test_lwe_concatenate(thread, axis, out_none):

    params = NuFHEParameters()
    lwe_params = params.in_out_params

    if axis == 0:
        shapes = [(3, 4), (1, 4), (4, 4)]
    elif axis == 1:
        shapes = [(4, 3), (4, 1), (4, 4)]

    ciphertexts = [mock_ciphertext(thread, lwe_params, shape) for shape in shapes]

    if out_none:
        out = None
    else:
        out = mock_ciphertext(thread, lwe_params, (8, 4) if axis == 0 else (4, 8))

    out = concatenate(ciphertexts, axis=axis, out=out)

    ref_a = numpy.concatenate([ciphertext.a.get() for ciphertext in ciphertexts], axis=axis)
    ref_b = numpy.concatenate([ciphertext.b.get() for ciphertext in ciphertexts], axis=axis)
    ref_cv = numpy.concatenate(
        [ciphertext.current_variances.get() for ciphertext in ciphertexts], axis=axis)

    assert (out.a.get() == ref_a).all()
    assert (out.b.get() == ref_b).all()
    assert numpy.allclose(out.current_variances.get(), ref_cv)


class _GetSlices:
    def __getitem__(self, index):
        return index

_get_slices = _GetSlices()

_lwe_assign_tests = [
    ((3, 4), _get_slices[1:], (3, 4), _get_slices[:-1], "[contig]=[contig]"),
    ((10,), _get_slices[1:10:2], (10,), _get_slices[:10:2], "[discontig]=[discontig]"),
    ((5,), _get_slices[1], (5,), _get_slices[2], "[scalar]=[scalar]"),
]

@pytest.mark.parametrize(
    'lwe_assign_test',
    [test[:-1] for test in _lwe_assign_tests],
    ids=[test[-1] for test in _lwe_assign_tests])
def test_lwe_assign(thread, lwe_assign_test):

    src_shape, src_slice, dst_shape, dst_slice = lwe_assign_test

    params = NuFHEParameters()
    lwe_params = params.in_out_params

    ct_src = mock_ciphertext(thread, lwe_params, src_shape)
    ct_dst = mock_ciphertext(thread, lwe_params, dst_shape)

    ref_a = ct_dst.a.get()
    ref_b = ct_dst.b.get()
    ref_cv = ct_dst.current_variances.get()

    ct_dst[dst_slice] = ct_src[src_slice]

    ref_a[dst_slice] = ct_src.a.get()[src_slice]
    ref_b[dst_slice] = ct_src.b.get()[src_slice]
    ref_cv[dst_slice] = ct_src.current_variances.get()[src_slice]

    assert (ct_dst.a.get() == ref_a).all()
    assert (ct_dst.b.get() == ref_b).all()
    assert numpy.allclose(ct_dst.current_variances.get(), ref_cv)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.core import Computation, Transformation, Parameter, Annotation, Type
import reikna.helpers as helpers

from .numeric_functions import Torus32, Int32
from .polynomial_transform import get_transform
from .performance import PerformanceParametersForDevice


TEMPLATE = helpers.template_for(__file__)


def get_tgsw_polynomial_decomp_trf(params: 'TGswParams', shape):
    tlwe_params = params.tlwe_params
    decomp_length = params.decomp_length
    mask_size = tlwe_params.mask_size
    polynomial_degree = tlwe_params.polynomial_degree

    result = Type(Int32, shape + (mask_size + 1, decomp_length, polynomial_degree))
    sample = Type(Torus32, shape + (mask_size + 1, polynomial_degree))
    return Transformation([
        Parameter('result', Annotation(result, 'o')),
        Parameter('sample', Annotation(sample, 'i'))],
        """
        <%
            mask = 2**params.bs_log2_base - 1
            half_base = 2**(params.bs_log2_base - 1)
        %>
        ${sample.ctype} sample = ${sample.load_idx}(${", ".join(idxs[:-2])}, ${idxs[-1]});
        int decomp_shift = 32 - (${idxs[-2]} + 1) * ${params.bs_log2_base};
        ${result.store_same}(
            (((sample + (${params.offset})) >> decomp_shift) & ${mask}) - ${half_base}
        );
        """,
        connectors=['results'],
        render_kwds=dict(params=params))


# result = result + p*sample
def get_tlwe_transformed_add_mul_to_trf(
        params: 'TGswParams', shape, bk_len: int, perf_params: PerformanceParametersForDevice):

    tlwe_params = params.tlwe_params
    decomp_length = params.decomp_length
    mask_size = tlwe_params.mask_size
    polynomial_degree = tlwe_params.polynomial_degree

    transform = get_transform(params.tlwe_params.transform_type)
    tdtype = transform.transformed_dtype()
    tlength = transform.transformed_length(polynomial_degree)
    tr_ctype = transform.transformed_internal_ctype()

    result = Type(tdtype, shape + (mask_size + 1, tlength))
    sample = Type(tdtype, shape + (mask_size + 1, decomp_length, tlength))
    bootstrap_key = Type(tdtype, (bk_len, mask_size + 1, decomp_length, mask_size + 1, tlength))

    return Transformation([
        Parameter('result', Annotation(result, 'o')),
        Parameter('sample', Annotation(sample, 'i')),
        Parameter('bootstrap_key', Annotation(bootstrap_key, 'i')),
        Parameter('bk_row_idx', Annotation(numpy.int32))],
        """
        ${tr_ctype} result = ${tr_ctype}pack(${dtypes.c_constant(0, result.dtype)});

        %for mask_idx in range(mask_size + 1):
        %for decomp_idx in range(decomp_length):
        {
            ${tr_ctype} a = ${tr_ctype}pack(
                ${sample.load_idx}(
                    ${", ".join(idxs[:-2])}, ${mask_idx}, ${decomp_idx}, ${idxs[-1]})
                );
            ${tr_ctype} b = ${tr_ctype}pack(
                ${bootstrap_key.load_idx}(
                    ${bk_row_idx}, ${mask_idx}, ${decomp_idx}, ${idxs[-2]}, ${idxs[-1]})
                );
            result = ${add}(result, ${mul_prepared}(a, b));
        }
        %endfor
        %endfor

        ${result.store_same}(${tr_ctype}unpack(result));
        """,
        connectors=['result'],
        render_kwds=dict(
            mask_size=mask_size,
            decomp_length=decomp_length,
            add=transform.transformed_add(perf_params),
            mul_prepared=transform.transformed_mul_prepared(perf_params),
            tr_ctype=tr_ctype))


class TGswTransformedExternalMul(Computation):

    def __init__(
            self, params: 'TGswParams', shape, bk_len, perf_params: PerformanceParametersForDevice):

        mask_size = params.tlwe_params.mask_size
        polynomial_degree = params.tlwe_params.polynomial_degree
        decomp_length = params.decomp_length

        transform = get_transform(params.tlwe_params.transform_type)
        tdtype = transform.transformed_dtype()
        tlength = transform.transformed_length(polynomial_degree)

        accum = Type(Torus32, shape + (mask_size + 1, polynomial_degree))
        bootstrap_key = Type(tdtype, (bk_len, mask_size + 1, decomp_length, mask_size + 1, tlength))

        self._params = params
        self._perf_params = perf_params
        self._shape = shape
        self._bk_len = bk_len

        Computation.__init__(self,
            [Parameter('accum', Annotation(accum, 'io')),
            Parameter('bootstrap_key', Annotation(bootstrap_key, 'i')),
            Parameter('bk_row_idx', Annotation(numpy.int32))])

    def _build_plan(self, plan_factory, device_params, accum, bootstrap_key, bk_row_idx):
        plan = plan_factory()

        perf_params = self._perf_params

        params = self._params

        tlwe_params = params.tlwe_params
        polynomial_degree = tlwe_params.polynomial_degree

        batch_shape = self._shape

        transform = get_transform(tlwe_params.transform_type)

        decomp = get_tgsw_polynomial_decomp_trf(params, batch_shape)
        decomp_and_ftr = transform.ForwardTransform(
            decomp.result.shape[:-1], polynomial_degree, perf_params)
        decomp_and_ftr.parameter.input.connect(decomp, decomp.result, sample=decomp.sample)

        add_mul = get_tlwe_transformed_add_mul_to_trf(
            params, batch_shape, self._bk_len, perf_params)
        add_mul_and_itr = transform.InverseTransform(
            add_mul.result.shape[:-1], polynomial_degree, perf_params)
        add_mul_and_itr.parameter.input.connect(
            add_mul, add_mul.result,
            tr_sample=add_mul.sample, bootstrap_key=add_mul.bootstrap_key,
            bk_row_idx=add_mul.bk_row_idx)

        tr_sample = plan.temp_array_like(decomp_and_ftr.parameter.output)

        plan.computation_call(decomp_and_ftr, tr_sample, accum)
        plan.computation_call(add_mul_and_itr, accum, tr_sample, bootstrap_key, bk_row_idx)

        return plan


class TGswAddMessage(Computation):

    def __init__(self, params: 'TGswParams', shape):

        self._params = params

        decomp_length = params.decomp_length
        mask_size = params.tlwe_params.mask_size
        polynomial_degree = params.tlwe_params.polynomial_degree

        result_a = Type(
            Torus32, shape + (mask_size + 1, decomp_length, mask_size + 1, polynomial_degree))
        messages = Type(Torus32, shape)

        Computation.__init__(self,
            [Parameter('result_a', Annotation(result_a, 'o')),
            Parameter('messages', Annotation(messages, 'i'))])

    def _build_plan(self, plan_factory, device_params, result_a, messages):
        plan = plan_factory()

        batch_len = helpers.product(messages.shape)
        plan.kernel_call(
            TEMPLATE.get_def("tgsw_add_message"),
            [result_a, messages],
            kernel_name="tgsw_add_message",
            global_size=(batch_len,),
            render_kwds=dict(
                slices=(len(messages.shape), 1, 1, 1, 1),
                base_powers=self._params.base_powers,
                decomp_length=self._params.decomp_length,
                mask_size=self._params.tlwe_params.mask_size))

        return plan
